#!/usr/bin/env python
#title           :pyvuka.py
#description     :This program is used for manipulating and fitting scientific
#                 data. Functions include global analysis, figure generation,
#                 and general statistical analysis of data.
#usage           :python pyvuka.py sys.argv[]
#python_version  :3.7
#==============================================================================
import sys
import os
from .ModuleLink import toPyVuka as pyvuka

__author__ = "R. Paul Nobrega"
__app_name__ = "PyVuka"
__description__ = "A General Purpose Global Data Analysis Package"
__copyright__ = "Copyright 2017, www.BostonAutoLytics.com"
__credits__ = "R. Paul Nobrega, Osman Bilsel, & David G. Lambright"
__license__ = "GPL"
__version__ = "0.0.259"
__maintainer__ = "R. Paul Nobrega"
__email__ = "Paul@BostonAutoLytics.com"
__status__ = "Development"


def homescreen():
    print("\n\n  ____        __     __       _     ")
    print(r" |  _ \  _   _\ \   / /_   _ | | __ __ _ ")
    print(r" | |_) || | | |\ \ / /| | | || |/ // _` |")
    print(r" |  __/ | |_| | \ V / | |_| ||   <| (_| |")
    print(r" |_|     \__, |  \_/   \__,_||_|\_\\__,_|")
    print(r"         |___/   ")
    print('\n'+__copyright__+'\n'+str(__credits__)+'\nVersion '+__version__+' ('+__status__+')')
    print("Questions or comments? Email: %s" % __email__)
    print("\n------------------------------------------")
    print("Type 'help' for information\nType 'q' to quit")
    print("------------------------------------------\n")
    return


def start():
    homescreen()
    exitflag = False
    pvk = pyvuka.initialize_instance()
    workingdir = os.getcwd()
    pvk.data.directories.working.set(workingdir)
    print("Current Working Directory: %s\n" % pvk.data.directories.working.get())
    if len(sys.argv[1:]) >= 1:
        if os.path.exists(sys.argv[1]):
            pvk.run_pyvuka_command("cwd -set "+str(sys.argv[1]))
            initargs = " ".join(map(str, sys.argv[2:])).split(';')
        else:
            initargs = " ".join(map(str, sys.argv[1:])).split(';')
        for initarg in initargs:
            print("PyVuka> " + initarg)
            print(pvk.run_pyvuka_command(initarg))
    while not exitflag:
        userinput = input("\nPyVuka> ")
        userinput = userinput.split(';')
        for command in userinput:
            output = pvk.run_pyvuka_command(command)
            if output == '':
                continue
            elif not output:
                exitflag = True
            else:
                print(output + "\n")


if __name__ == '__main__':
    start()



import numpy as np
import xlrd
import xlsxwriter as XL
import xml.etree.ElementTree as xmlio
import base64
from . import plot, fitfxns, inputprocessing
from matplotlib import pyplot as pl
import os
import chardet
import array
import string
from matplotlib import rcParams
import io
from io import BytesIO as BIO
rcParams.update({'figure.autolayout': True})


class IO(object):
    def __init__(self, data_instance):
        self.data = data_instance.data

    def writepvk(self, filename):
        return "File type does not exist yet!"

    def readpvk(self, filename):
        #filetype does not exist yet
        return "File type does not exist yet!"

    def readsvb(self, filename, filestruct):
        return "File read fxn not written yet!"

    def readtxt(self, filename, filestruct, comparams):
        print("Copying text file to memory (process is slow with large files)...\n")
        in_txt = []
        row = "temp"
        col = "temp"
        numstruct = "temp"
        numlines = "temp"
        valerror = "Non-numeric input! Process interrupted!"
        structwidth = 0
        onexcol = False
        xindex = None
        xeindex = None
        yindex = None
        yeindex = None
        zindex = None
        zeindex = None
        iscat = False

        delim = detect_delimiter(filename)

        with open(filename) as textFile:
            for line in textFile:
                in_txt.append(line.split(delim))

        if filestruct == "-y":
            structwidth = 1
            onexcol = True
            yindex = 0
        if filestruct == "-xy" or filestruct == "-ye":
            structwidth = 2
            xindex = 0
            yindex = 1
            if filestruct == "-ye":
                onexcol = True
                xindex = None
                yindex = 0
                yeindex = 1
        elif filestruct == "-xyz" or filestruct == "-xye" or filestruct == "-cye" \
                or filestruct == "-cyz" or filestruct == "-ccz":
            structwidth = 3
            if filestruct == "-xyz":
                xindex = 0
                yindex = 1
                zindex = 2
            elif filestruct == "-xye":
                xindex = 0
                yindex = 1
                yeindex = 2
            elif filestruct == "-cye":
                structwidth = 3
                iscat = True
                xindex = 0
                yindex = 1
                yeindex = 2
            elif filestruct == "-cyz":
                structwidth = 3
                iscat = True
                xindex = 0
                yindex = 1
                zindex = 2
            elif filestruct == "-ccz":
                structwidth = 3
                iscat = True
                xindex = 0
                yindex = 1
                zindex = 2
        elif filestruct == "-xeye":
            structwidth = 4
            xindex = 0
            xeindex = 1
            yindex = 2
            yeindex = 3
        elif filestruct == "-xeyeze":
            structwidth = 6
            xindex = 0
            xeindex = 1
            yindex = 2
            yeindex = 3
            zindex = 4
            zeindex = 5
        inparse = inputprocessing.InputParser()
        inparse.prompt = ["Number of rows from top to skip", "Number of columns from left to skip",
                  "Number of datasets to read (0 is until end of file)", "Number of lines to read (0 is until end of file)"]
        inparse.inputbounds = [[0,1E100],[0,1E100],[0,1E100],[0,1E100]]
        inparse.defaultinput = ['0', '0', '0', '0']
        inparse.userinput = comparams
        row, col, numstruct, numlines = [int(x) for x in inparse.getparams()]
        inparse.userinput = comparams
        if numstruct <= 0:
            numstruct = int(int(len(in_txt[row]) - col)/structwidth)
        if numlines <= 0:
            numlines = int(len(in_txt) - row)
        count = 0
        commonx = []
        maxrow = int(numlines+row)
        maxcol = int((numstruct*structwidth)+col)
        if onexcol:
            for rows in range(row, maxrow):
                commonx.append(float(in_txt[rows][col]))
            col = col+1
        while (col+count < maxcol):
            newbuffer = self.data.new_buffer()
            print("Last column read: " + str(count+structwidth) + " of " + str(maxcol))
            skipped = 0
            for rows in range(row, maxrow):
                valindex = 0
                for cols in range(col + count, col + count + structwidth):
                    if xindex is not None and valindex == xindex:
                        xtemp = in_txt[rows][cols]
                        if xtemp == '':
                            skipped += 1
                        else:
                            try:
                                newbuffer.data.x.append(float(xtemp))
                            except ValueError:
                                return valerror
                    if xeindex is not None and valindex == xeindex:
                        xetemp = in_txt[rows][cols]
                        if xetemp == '':
                            skipped += 1
                        else:
                            try:
                                newbuffer.data.xe.append(float(xetemp))
                            except ValueError:
                                return valerror
                    if yindex is not None and valindex == yindex:
                        ytemp = in_txt[rows][cols]
                        if ytemp == '':
                            skipped += 1
                        else:
                            try:
                                newbuffer.data.y.append(float(ytemp))
                            except ValueError:
                                return valerror
                    if yeindex is not None and valindex == yeindex:
                        yetemp = in_txt[rows][cols]
                        if yetemp == '':
                            skipped += 1
                        else:
                            try:
                                newbuffer.data.ye.append(float(yetemp))
                            except ValueError:
                                return valerror
                    if zindex is not None and valindex == zindex:
                        ztemp = in_txt[rows][cols]
                        if ztemp == '':
                            skipped += 1
                        else:
                            try:
                                newbuffer.data.z.append(float(ztemp))
                            except ValueError:
                                return valerror
                    if zeindex is not None and valindex == zeindex:
                        zetemp = in_txt[rows][cols]
                        if zetemp == '':
                            skipped += 1
                        else:
                            try:
                                newbuffer.data.ze.append(float(zetemp))
                            except ValueError:
                                return valerror
                    if skipped >= 8:
                        break
                    valindex += 1
            count += structwidth
            print(f"\tLines read into buffer: {newbuffer.data.x.length()}")
            if onexcol:
                newbuffer.data.x.set(commonx)
            if iscat:
                newbuffer.category.x.set([str(x) for x in newbuffer.data.x.get()])
                newbuffer.data.x.clear()
            if iscat and filestruct == "-ccz":
                newbuffer.category.y.set([str(x) for x in newbuffer.data.y.get()])
                newbuffer.data.y.clear()
            self.data.matrix.add_buffer(newbuffer)
        self.colorallseries()
        self.buffertoseries()
        self.seriestotitle()
        return "Data has been read into memory."

    def readxlsx(self, filename, filestruct, comparams):
        print("Copying XLSX file to memory (process is slow with large files)...\n")
        in_xlsx = xlrd.open_workbook(filename)
        print("XLSX Sheet Names:")
        row = "temp"
        col = "temp"
        sheet = "temp"
        numstruct = "temp"
        numlines = "temp"
        valerror = "Non-numeric input! Process interrupted!"
        structwidth = 0
        onexcol = False
        xindex = None
        xeindex = None
        yindex = None
        yeindex = None
        zindex = None
        zeindex = None
        iscat = False

        if filestruct == "-y":
            structwidth = 1
            onexcol = True
            yindex = 0
        if filestruct == "-xy" or filestruct == "-ye":
            structwidth = 2
            xindex = 0
            yindex = 1
            if filestruct == "-ye":
                onexcol = True
                xindex = None
                yindex = 0
                yeindex = 1
        elif filestruct == "-xyz" or filestruct == "-xye" or filestruct == "-cye" \
                or filestruct == "-cyz" or filestruct == "-ccz":
            structwidth = 3
            if filestruct == "-xyz":
                xindex = 0
                yindex = 1
                zindex = 2
            elif filestruct == "-xye":
                xindex = 0
                yindex = 1
                yeindex = 2
            elif filestruct == "-cye":
                structwidth = 3
                iscat = True
                xindex = 0
                yindex = 1
                yeindex = 2
            elif filestruct == "-cyz":
                structwidth = 3
                iscat = True
                xindex = 0
                yindex = 1
                zindex = 2
            elif filestruct == "-ccz":
                structwidth = 3
                iscat = True
                xindex = 0
                yindex = 1
                zindex = 2
        elif filestruct == "-xeye":
            structwidth = 4
            xindex = 0
            xeindex = 1
            yindex = 2
            yeindex = 3
        elif filestruct == "-xeyeze":
            structwidth = 6
            xindex = 0
            xeindex = 1
            yindex = 2
            yeindex = 3
            zindex = 4
            zeindex = 5
        inparse = inputprocessing.InputParser()
        inparse.prompt = ["Sheet to read in", "Number of rows from top to skip", "Number of columns from left to skip",
                  "Number of datasets to read (0 is until end of file)", "Number of lines to read (0 is until end of file)"]
        inparse.inputbounds = [[0,1E100],[0,1E100],[0,1E100],[0,1E100],[0,1E100]]
        inparse.defaultinput = ['0', '0', '0', '0', '0']
        inparse.userinput = comparams
        sheet_names = in_xlsx.sheet_names()
        if len(sheet_names) == 1:
            if len(inparse.userinput) > 0:
                inparse.userinput[0] = '0'
            else:
                inparse.userinput.append('0')
        else:
            for name in range(len(sheet_names)):
                print("\t[" + str(name) + "] " + sheet_names[name])
            print("\n")
        sheet, row, col, numstruct, numlines = [int(x) for x in inparse.getparams()]
        inparse.userinput = comparams
        if numstruct <=0:
            numstruct = int(int(in_xlsx.sheet_by_index(sheet).ncols - col)/structwidth)
        if numlines <= 0:
            numlines = int(in_xlsx.sheet_by_index(sheet).nrows - row)
        count = 0
        commonx = []
        maxrow = int(numlines+row)
        maxcol = int((numstruct*structwidth)+col)
        if onexcol:
            for rows in range(row, maxrow):
                commonx.append(float(in_xlsx.sheet_by_index(sheet).cell_value(rows, col)))
            col = col+1
        while (col+count < maxcol):
            newbuffer = self.data.new_buffer()
            print("Last column read: " + str(count+structwidth) + " of " + str(maxcol))
            skipped = 0
            for rows in range(row, maxrow):
                valindex = 0
                for cols in range(col + count, col + count + structwidth):
                    if xindex is not None and valindex == xindex:
                        xtemp = in_xlsx.sheet_by_index(sheet).cell_value(rows, cols)
                        if xtemp == '':
                            skipped += 1
                        else:
                            try:
                                newbuffer.data.x.append(float(xtemp))
                            except ValueError:
                                return valerror
                    if xeindex is not None and valindex == xeindex:
                        xetemp = in_xlsx.sheet_by_index(sheet).cell_value(rows, cols)
                        if xetemp == '':
                            skipped += 1
                        else:
                            try:
                                newbuffer.data.xe.append(float(xetemp))
                            except ValueError:
                                return valerror
                    if yindex is not None and valindex == yindex:
                        ytemp = in_xlsx.sheet_by_index(sheet).cell_value(rows, cols)
                        if ytemp == '':
                            skipped += 1
                        else:
                            try:
                                newbuffer.data.y.append(float(ytemp))
                            except ValueError:
                                return valerror
                    if yeindex is not None and valindex == yeindex:
                        yetemp = in_xlsx.sheet_by_index(sheet).cell_value(rows, cols)
                        if yetemp == '':
                            skipped += 1
                        else:
                            try:
                                newbuffer.data.ye.append(float(yetemp))
                            except ValueError:
                                return valerror
                    if zindex is not None and valindex == zindex:
                        ztemp = in_xlsx.sheet_by_index(sheet).cell_value(rows, cols)
                        if ztemp == '':
                            skipped += 1
                        else:
                            try:
                                newbuffer.data.z.append(float(ztemp))
                            except ValueError:
                                return valerror
                    if zeindex is not None and valindex == zeindex:
                        zetemp = in_xlsx.sheet_by_index(sheet).cell_value(rows, cols)
                        if zetemp == '':
                            skipped += 1
                        else:
                            try:
                                newbuffer.data.ze.append(float(zetemp))
                            except ValueError:
                                return valerror
                    if skipped == 6:
                        break
                    valindex += 1
            count += structwidth
            t = newbuffer.data.x.length()
            print(f"\tLines read into buffer: {newbuffer.data.x.length()}")
            if onexcol:
                newbuffer.data.x.set(commonx)
            if iscat:
                newbuffer.category.x.set([str(x) for x in newbuffer.data.x.get()])
                newbuffer.data.x.clear()
            if iscat and filestruct == "-ccz":
                newbuffer.category.y.set([str(x) for x in newbuffer.data.y.get()])
                newbuffer.data.y.clear()
            self.data.matrix.add_buffer(newbuffer)
        self.colorallseries()
        self.buffertoseries()
        self.seriestotitle()
        return "Data has been read into memory."

    def readfb(self, experimentdirectory):
        if not os.path.exists(experimentdirectory):
            return False
        experiment = []
        for files in os.listdir(experimentdirectory):
            if files.endswith('.frd') and not files.startswith('._'):
                Xdata = []
                Ydata = []
                StepName = []
                ActualTime = []
                StepStatus = []
                StepType = []
                Concentration = []
                MolarConcentration = []
                SampleID = []
                WellType = []
                MW = []
                Flags =[]
                SampleGroup = []
                StepLoc = []
                loadingsample = ''
                loadingstart = 0
                loadingend = 0
                loadingwell =''
                infile = xmlio.parse(os.path.join(experimentdirectory, files)).getroot()
                for expinfo in infile.findall('ExperimentInfo'):
                    SensorName = expinfo.find('SensorName').text
                    SensorType = expinfo.find('SensorType').text
                    SensorRole = expinfo.find('SensorRole').text
                    SensorInfo = expinfo.find('SensorInfo').text
                for kindata in infile.findall('KineticsData'):
                    for stepdata in kindata.findall('Step'):
                        for commondata in stepdata.findall('CommonData'):

                            WellType.append(commondata.find('WellType').text)
                            Concentration.append(commondata.find('Concentration').text)
                            MolarConcentration.append(commondata.find('MolarConcentration').text)
                            SampleID.append(commondata.find('SampleID').text)
                            if commondata.find('SampleGroup') is not None:
                                SampleGroup.append(commondata.find('SampleGroup').text)
                            else:
                                SampleGroup.append(None)
                            MW.append(commondata.find('MolecularWeight').text)
                            Xdata.append(np.array(array.array('f', base64.b64decode(stepdata.find('AssayXData').text))))
                            Ydata.append(np.array(array.array('f', base64.b64decode(stepdata.find('AssayYData').text))))
                            StepName.append(stepdata.find('StepName').text)
                            ActualTime.append(stepdata.find('ActualTime').text)
                            StepStatus.append(stepdata.find('StepStatus').text)
                            StepLoc.append(commondata.find('SampleRow').text + commondata.find('SampleLocation').text)
                            StepType.append(stepdata.find('StepType').text)
                            if StepType[-1].upper() == 'LOADING' and SampleID[-1] is not None:
                                loadingsample = SampleID[-1]
                                loadingstart = Ydata[-1][0]
                                loadingend = Ydata[-1][-1]
                                loadingwell = StepLoc[-1]
                for status in StepStatus:
                    if not status == 'OK':
                        Flags.append('Sensor:' + status)
                        break

                newbuffer = self.data.new_buffer()

                ### copy and paste of 'interstepcorrection' method from historical fortebiopkg. edited to suit
                for j in range(0, len(Ydata) - 1):
                    if Ydata[j][-1] > Ydata[j + 1][0]:
                        ydif = Ydata[j][-1] - Ydata[j + 1][0]
                        for k in range(0, len(Ydata[j + 1])):
                            Ydata[j + 1][k] += ydif
                    else:
                        ydif = Ydata[j + 1][0] - Ydata[j][-1]
                        for k in range(0, len(Ydata[j + 1])):
                            Ydata[j + 1][k] -= ydif
                for j in range(0, len(Xdata) - 1):
                    if Xdata[j][-1] < Xdata[j + 1][0]:
                        xdif = Xdata[j + 1][0] - Xdata[j][-1]
                        for k in range(0, len(Xdata[j + 1])):
                            Xdata[j + 1][k] -= xdif
                    else:
                        xdif = Xdata[j][-1] - Xdata[j + 1][0]
                        for k in range(0, len(Xdata[j + 1])):
                            Xdata[j + 1][k] -= xdif

                for segment in Xdata:
                    newbuffer.data.x.append(segment)
                    newbuffer.plot.axis.x.lines.append(segment[0])

                newbuffer.data.x.set(np.concatenate(Xdata, axis=None))
                newbuffer.data.y.set(np.concatenate(Ydata, axis=None))
                ###
                newbuffer.data.z.set([float(MolarConcentration[-2])] * newbuffer.data.y.length())

                newbuffer.comments.set([str(SensorInfo) + " on " + str(SensorType) + " vs " +
                                           str(SampleID[-2]) + " @ " + str(MolarConcentration[-2]) + "nM"])
                newbuffer.plot.series.name.set(newbuffer.comments.get())
                newbuffer.plot.title.set(newbuffer.comments.get())
                newbuffer.plot.axis.x.title.set("Time (s)")
                newbuffer.plot.axis.y.title.set("Response (nm)")
                newbuffer.plot.axis.x.lines.show()
                newbuffer.meta_dict = {'xData': Xdata, 'yData': Ydata,
                                       'stepName': StepName, 'actualTime': ActualTime, 'sensorType': SensorType,
                                       'stepStatus': StepStatus, 'stepType': StepType, 'concentration': Concentration,
                                       'molarConcentration':MolarConcentration, 'sampleID':SampleID,
                                       'wellType': WellType, 'mw': MW, 'flags':Flags, 'sampleGroup': SampleGroup,
                                       'stepLocation': StepLoc, 'loadingSample': loadingsample, 'sensorInfo': SensorInfo,
                                       'loadingStart':loadingstart, 'loadingEnd':loadingend, 'loadingWell':loadingwell,
                                       'inFile':infile, 'sensorName':SensorName}
                self.data.matrix.add_buffer(newbuffer)
        self.colorallseries()
        return "ForteBio data read into memory."

    def readi3x(self, experimentdirectory):
        if not os.path.exists(experimentdirectory):
            return False
        experiment = {"Well": [], "Sample_ID": [], "Y_Signal": [], "Concentration": []}
        filecount = 0
        for files in os.listdir(experimentdirectory):
            if files.endswith('.txt') and ('[' in files and ']' in files):
                filename = os.path.join(experimentdirectory, files)
                concentration = filename.split('[')[1].split(']')[0]
                if concentration[-2:] == "ng":
                    concentration = float(concentration[:-2])*1E-9
                elif concentration[-2:] == "ug":
                    concentration = float(concentration[:-2])*1E-6
                elif concentration[-2:] == "mg":
                    concentration = float(concentration[:-2])*1E-3
                with open(filename) as rawdata:
                    datalines = rawdata.readlines()
                    readdata = False
                    for line in datalines:
                        line = filter(lambda x: x in string.printable, line)
                        if "Sample\tWells\tValue" in line:
                            readdata = True
                        if readdata and "Sample\tWells\tValue" not in line:
                            if "Group Column\tFormula Name" in line or line == '\r\n':
                                break
                            if filecount == 0:
                                temp = line.split('\t')
                                experiment["Sample_ID"].append(temp[0])
                                experiment["Well"].append(temp[1])
                                experiment["Y_Signal"].append([float(temp[2])])
                                experiment["Concentration"].append([concentration])
                            else:
                                temp = line.split('\t')
                                wellindex = experiment["Well"].index(temp[1])
                                experiment["Y_Signal"][wellindex].append(float(temp[2]))
                                experiment["Concentration"][wellindex].append(concentration)
                filecount += 1
        for i in range(len(experiment["Well"])):
            newbuffer = self.data.new_buffer()
            experiment["Concentration"][i], experiment["Y_Signal"][i] = (list(t) for t in zip(*sorted(zip(experiment["Concentration"][i], experiment["Y_Signal"][i]))))
            newbuffer.data.x.set(experiment["Concentration"][i])
            newbuffer.data.y.set(experiment["Y_Signal"][i])
            newbuffer.plot.series.name.set(experiment["Sample_ID"][i] + " (Well: " + experiment["Well"][i] + ")")
            newbuffer.comments.add("Well: "+experiment["Well"][i])
            newbuffer.comments.add("Sample_ID: " + experiment["Sample_ID"][i])
            self.data.matrix.add_buffer(newbuffer)
        self.colorallseries()
        self.seriestotitle()
        return "Data has been read into memory."

    def writexlsx(self, filename, **kwargs):
        xlsx = xlsx_out(filename)
        xlsx.sheet_name = kwargs.get("sheet_name")
        xlsx.add_header(kwargs.get("header_list"))
        xlsx.set_col_widths(kwargs.get("col_width_list"))
        xlsx.set_row_heights(kwargs.get("row_heights"))

        plot_out = plot.plotter()

        fitter = fitfxns.datafit()
        fitter.funcindex = self.data.matrix.buffer(1).fit.function_index.get()
        fitter.applyfxns()

        if not xlsx.header:
            auto_header = ['Sample']
            for val in fitter.paramid:
                auto_header.extend([val, str(val + '_error')])
            auto_header.extend(['xsq', 'rsq', 'plot'])
            xlsx.add_header(auto_header)

        if not xlsx.row_heights:
            xlsx.set_row_heights(152)

        if not xlsx.col_widths:
            xlsx.set_col_widths([20] * len(xlsx.header))

        if not xlsx.sheet_name:
            xlsx.sheet_name = "Output"

        for i in range(1, self.data.matrix.length() + 1):
            xl_line = []
            buffer = self.data.matrix.buffer(i)
            xl_line.extend([buffer.plot.series.name.get()])
            for j in range(len(buffer.fit.parameter.get())):
                xl_line.extend([buffer.fit.parameter.get()[j]])
                xl_line.extend([buffer.fit.parameter_error.get()[j]])
            xl_line.append(buffer.fit.chisq.get())
            xl_line.append(buffer.fit.rsq.get())
            xl_line.append(BIO(plot_out([i], get_bytes=True)))
            xlsx.add_line(xl_line)
        xlsx.write_xlsx()

        for i in range(1, self.data.matrix.length() + 1):
            xlsx = xlsx_out(filename[:-5] + f"_sensor{i}_raw.xlsx")
            xlsx.sheet_name = "Output"
            xlsx.add_line([buffer.plot.series.name.get(), 'X', 'Y', 'X-breaks']) #header
            buffer = self.data.matrix.buffer(i)
            tbufx = buffer.data.x.get()
            tbufy = buffer.data.y.get()
            tbreakx = buffer.plot.axis.x.lines.get()
            for j in range(len(tbufx)):
                if j < len(tbreakx):
                    xlsx.add_line(['', tbufx[j], tbufy[j], tbreakx[j]])
                else:
                    xlsx.add_line(['', tbufx[j], tbufy[j]])
            xlsx.write_xlsx()
        return

    def colorallseries(self):
        start = 0.0
        stop = 1.0
        number_of_buffers = self.data.matrix.length()
        cm_subsection = np.linspace(start, stop, number_of_buffers)

        colors = [pl.cm.gist_rainbow(x) for x in cm_subsection]

        for i, color in enumerate(colors):
            self.data.matrix.buffer(i).plot.series.color.set(color)
        return True

    def buffertoseries(self):
        for i in range(self.data.matrix.length()):
            self.data.matrix.buffer(i).plot.series.name.set(f"Buffer {i + 1}")
        return True

    def seriestotitle(self):
        for i in range(self.data.matrix.length()):
            self.data.matrix.buffer(i).plot.title.set(self.data.matrix.buffer(i).plot.series.name.get())
        return True


class xlsx_out(object):
    def __init__(self, filename):
        """Return an xlsx object whose name is filename """
        self.filename = filename
        self.matrix = []
        self.header = []
        self.col_widths = []
        self.row_heights = []
        self.sheet_name = 'output'

    def add_line(self, write_array):
        self.matrix.append(write_array)
        return

    def add_header(self, write_array):
        self.header = write_array
        return

    def set_col_widths(self, widths):
        if not widths:
            return
        maxcol = -10
        if ',' in str(widths):
            array_out = list(widths)
        else:
            if not self.header:
                self.col_widths = widths
                return
            if len(self.header) > maxcol:
                maxcol = len(self.header)
            for line in self.matrix:
                if len(line) > maxcol:
                    maxcol = len(line)
            array_out = [widths for j in range(maxcol)]
        self.col_widths = array_out
        return

    def set_row_heights(self, heights):
        if not self.header:
            header = 0
        else:
            header = 1
        if not heights:
            return
        maxrow = -10
        if ',' in str(heights):
            array_out = list(heights)
        else:
            if not self.matrix:
                self.row_heights = heights
                return
            elif len(self.matrix) > maxrow:
                maxrow = len(self.matrix)
            array_out = [heights for j in range(maxrow + header)]
        self.row_heights = array_out
        return

    def write_xlsx(self):
        self.set_row_heights(self.row_heights)
        self.set_col_widths(self.col_widths)
        wb = XL.Workbook(self.filename)
        ws = []
        ws.append(wb.add_worksheet(self.sheet_name))
        header = 0

        if self.header:
            for i in range(len(self.header)):
                ws[0].write(0, i, self.header[i])
            ws[0].freeze_panes(1, 0)
            header = 1

        for i in range(len(self.matrix)):
            for j in range(len(self.matrix[i])):
                var_type = str(type(self.matrix[i][j]))
                if '_io.BytesIO' in var_type:
                    ws[0].insert_image(i + header, j, 'figure.png',
                                       {'image_data': self.matrix[i][j], 'x_scale': 1, 'y_scale': 1})
                else:
                    ws[0].write(i + header, j, self.matrix[i][j])

        if self.col_widths:
            for i in range(len(self.col_widths)):
                ws[0].set_column(i, i, self.col_widths[i])

        if self.row_heights:
            for i in range(len(self.row_heights)):
                ws[0].set_row(i, self.row_heights[i])
            if self.header:
                ws[0].set_row(0, 18)

        wb.close()
        return


def detect_delimiter(filename):
    '''Determine if comma or tab delimited'''
    encode_type = predict_encoding(filename)
    with io.open(filename, 'r', encoding=encode_type) as textFile:
        linein = textFile.readline()
        if len(linein.split(',')) < len(linein.split('\t')):
            return '\t'
        else:
            return ','


def predict_encoding(file_path, n_lines=20):
    '''Predict a file's encoding using chardet'''
    # Open the file as binary data
    with open(file_path, 'rb') as f:
        # Join binary lines for specified number of lines
        rawdata = b''.join([f.readline() for _ in range(n_lines)])
    return chardet.detect(rawdata)['encoding']




import sys
import os.path
from . import plot, fitfxns, fileio, Modules, numericalmethods, inputprocessing
import math
import numpy as np
import copy


class Command(object):
    """extend with do_something  method to handle your commands"""

    def __init__(self, data_instance, quit_commands=('q', 'quit', 'exit'), help_commands=('help', '?', 'h')):
        self._quit_cmd = quit_commands
        self._help_cmd = help_commands
        self.inst = data_instance
        self.fitter = fitfxns.datafit(data_instance)

    def __call__(self, line):
        if line is None or line == '':
            return ''
        cmd, args = self.__parse_cmd_line(line)
        safe_commands = ['read', 'rea', 'exec', 'cwd']
        if cmd in self._quit_cmd:
            return False
        elif cmd in self._help_cmd:
            return self.help(args[0] if args else None)
        elif cmd not in safe_commands and self.inst.data.matrix.get() == []:
            return "\n No Data In Memory! Read data with command: rea ."
        elif hasattr(self, 'do_' + cmd):
            return getattr(self, 'do_' + cmd)(*args)
        elif hasattr(self, 'do_' + cmd[:4]):
            return getattr(self, 'do_' + cmd[:4])(*args)
        elif hasattr(self, 'do_' + cmd[:3]):
            return getattr(self, 'do_' + cmd[:3])(*args)
        elif hasattr(self, 'do_' + cmd[:2]):
            return getattr(self, 'do_' + cmd[:2])(*args)
        else:
            return "\nUnknown Command: %s\n" % cmd

    def __parse_cmd_line(self, line):
        '''Parses commandline input.  line is space delimited except when double quotes are used'''
        dbl_qoute_idicies = [pos for pos, char in enumerate(line) if char == '"']
        dbl_qoute_bounds = []
        white_spaces = ['', ' ', '  ', '"', ' " ', ' "', ' "']
        tokens = []
        args=[]

        if len(dbl_qoute_idicies) > 1:
            for x in range(0, int(math.floor(len(dbl_qoute_idicies)/2),), 2):
                dbl_qoute_bounds.append([dbl_qoute_idicies[x], dbl_qoute_idicies[x+1]])

            min_bound = 0
            for bound in dbl_qoute_bounds:
                max_bound = bound[0]
                temp = line[min_bound:max_bound].split(' ')
                tokens.extend(temp)
                tokens.append(line[bound[0]:bound[1]].replace('"', ""))
                min_bound = bound[1]
            if min_bound < len(line):
                tokens.extend(line[min_bound:len(line)].split(' '))
        else:
            tokens = line.split(' ')

        tokens = [seg for seg in tokens if seg not in white_spaces]

        cmd = tokens[0].lower()
        if len(tokens) > 1:
            args = tokens[1:]
        return cmd, args

    def do_pl(self, *args):
        """\nCommand: Plot Limits\n
        Description: Specifies global x range of data to be used for drawing, fitting, writing, and manipulating data.

        Example Usage:
        \tpl on                  (Makes Plot Limits active)
        \tpl off                 (Disables Plot Limits)
        \tpl 1 10 20 100         (Plot limits = buffers 1-10, x range 20-100 by point index)
        \tpl -v 1 10 5.36 2000.1 (Plot limits = buffers 1-10, x range 5.36-2000.1 by x value)

        Default Inputs: on [first buffer] [last buffer] [x min of shortest buffer] [x max of shortest buffer]

        Default Options: N/A

        Options:
        \t-v option uses x values instead of point indicies to set limits
        """

        inparse = inputprocessing.InputParser()
        lastbuffer = self.inst.data.matrix.length()
        shortestxdata = 0
        if self.inst.data.matrix.length() == 0:
            return "No Data in Matrix!"
        if ((len(args) == 1 and args[0].lower() in ['on','-v']) or len(args) == 0) and self.inst.data.plot_limits.is_active:
            return "Plot Limits Are Already Active"
        elif len(args) == 1 and args[0].lower() == 'off' and not self.inst.data.plot_limits.is_active:
            return "Plot Limits Are Already Inactive"
        if (len(args) > 0 and args[0].lower() == 'off') and self.inst.data.plot_limits.is_active:
            self.inst.data.plot_limits.off()
            return "Plot Limits Off!"

        min_x_val = min([self.inst.data.matrix.buffer(i).data.x.min() for i in range(1, self.inst.data.matrix.length() + 1)])
        max_x_val = max([self.inst.data.matrix.buffer(i).data.x.max() for i in range(1, self.inst.data.matrix.length() + 1)])

        if "-v" in args:
            inparse.prompt = ["First Buffer", "Last Buffer", "Min x Val", "Max x Val"]
            inparse.inputbounds = [[1, lastbuffer], [1, lastbuffer], [-1E100, 1E100], [-1E100, 1E100]]
            inparse.defaultinput = ['1', str(lastbuffer), str(min_x_val), str(max_x_val)]
        else:
            min_idx = self.inst.data.matrix.buffer(1).data.x.nearest_index_to_value(min_x_val) + 1
            max_idx = self.inst.data.matrix.buffer(1).data.x.nearest_index_to_value(max_x_val) + 1
            inparse.prompt = ["First Buffer", "Last Buffer", "First Point", "Last Point"]
            inparse.inputbounds = [[1, lastbuffer], [1, lastbuffer],
                                   [1, min_idx],
                                   [1, max_idx]]
            inparse.defaultinput = ['1', str(lastbuffer), str(min_idx), str(max_idx)]
        if not inparse(args):
            return "Invalid input!"
        else:
            if not inparse.getparams():
                return "\nNo changes Have Been Made to Plot Limits!"
            if 'on' or '-on' in inparse.modifiers and not self.inst.data.plot_limits.is_active:
                brange = sorted([int(inparse.userinput[0]), int(inparse.userinput[1])])
                self.inst.data.plot_limits.buffer_range.set(brange)
                minindex = []
                maxindex = []
                if '-v' in inparse.cmdflags:
                    min_x = float(inparse.userinput[2])
                    max_x = float(inparse.userinput[3])
                    for i in range(min(brange), max(brange) + 1):
                        minindex.append(self.inst.data.matrix.buffer(i).data.x.nearest_index_to_value(min_x))
                        maxindex.append(self.inst.data.matrix.buffer(i).data.x.nearest_index_to_value(max_x))
                    self.inst.data.plot_limits.x_range.set([min(minindex) + 1, min(maxindex) + 1])
                else:
                    pointnummin=int(inparse.userinput[2])
                    pointnummax=int(inparse.userinput[3])
                    self.inst.data.plot_limits.x_range.set([pointnummin, pointnummax])
                self.inst.data.plot_limits.on()

        if not self.inst.data.plot_limits.is_active and 'off' or '-off' in inparse.modifiers:
            self.inst.data.plot_limits.off()
            return "Plot Limits Off!"
        elif self.inst.data.plot_limits.is_active == True:
            return "Plot Limits On!"
        else:
            return "Error Configuring Plot Limits!"

    def do_rea(self, *args):
        """\nCommand: REAd\n
        Description: Read in data from various file types. Horizontally repeating units will be read. Each data block will be saved in memory as one data buffer

        Example Usage:
        \tread -txt -xy c:\data\data.txt (Read specified ascii test file with xy data structure.  Structure definitions will be prompted)
        \tread -xlsx -xy c:\data\data.txt 0 3 0 0 0 (Read specified xlsx file with xy data structure, 3 row header, and default structure options)

        Default Input: N/A

        Default Options: -txt -xy  (if no file format given and extension matches a format, the extension will be used)

        Usage Options:
        \t-cwd  (uses cwd as base directory [used for for multiple file formats only])

        File Format Options:
        \t-txt (ascii text, tab or comma delimited)
        \t-xlsx (excel xlsx)
        \t-svb (legacy savuka binary)
        \t-fb (fortebio raw files)
        \t-pvk (pyvuka xml file)
        \t-i3x (SpectraMax i3x text file)

        \tData Block Configuration Options:
        \t-y  (first column is common x to single or repeating y columns.)
        \t-xy
        \t-xyz
        \t-ye
        \t-xye
        \t-xeye
        \t-xeyeze
        \t-cye  (First column is category (non-numeric), second is Y-val, third is Y-Error.)
        \t-cyz  (First column is category (non-numeric), second is Y-val, third is Z-val.)
        \t-ccz  (First column is category (non-numeric), second is category (non-numeric), third is Z-val.)
        """
        inparse = inputprocessing.InputParser()
        if not inparse(args):
            return "Invalid input!"
        formatoptions = ["txt", "xlsx", "svb", "fb", "pvk", "i3x"]
        userflags = inparse.cmdflags
        comparams = inparse.userinput

        if "-cwd" in userflags:
            if not inparse.modifiers:
                inparse.modifiers = [self.inst.data.directories.working.get()]
            else:
                inparse.modifiers[0] = self.inst.data.directories.working.get()

        if len(inparse.modifiers) == 0:
            return "No input file specified!"
        else:
            filetoread = inparse.modifiers[0]
            if not os.path.exists(filetoread):
                return "Invalid Filename!"
        try:
            ext = filetoread.split('.')[-1]
            if ext.lower() == 'csv':
                ext = 'txt'
            if ext in formatoptions and ('-' + ext) not in userflags:
                userflags.append('-' + ext)
        except:
            ext = None
        if len(userflags) == 1:
            userflags.append("-xy")
        try:
            fio = fileio.IO(self.inst)
            if "-txt" in userflags:
                if "-y" in userflags:
                    return fio.readtxt(filetoread, "-y", comparams)
                elif "-xy" in userflags:
                    return fio.readtxt(filetoread, "-xy", comparams)
                elif "-xyz" in userflags:
                    return fio.readtxt(filetoread, "-xyz", comparams)
                elif "-ye" in userflags:
                    return fio.readtxt(filetoread, "-ye", comparams)
                elif "-xye" in userflags:
                    return fio.readtxt(filetoread, "-xye", comparams)
                elif "-xeye" in userflags:
                    return fio.readtxt(filetoread, "-xeye", comparams)
                elif "-xeyeze" in userflags:
                    return fio.readtxt(filetoread, "-xeyeze", comparams)
                elif "-cye" in userflags:
                    return fio.readtxt(filetoread, "-cye", comparams)
                elif "-cyz" in userflags:
                    return fio.readtxt(filetoread, "-cyz", comparams)
                elif "-ccz" in userflags:
                    return fio.readtxt(filetoread, "-ccz", comparams)
                else:
                    return "No txt file structure given."
            elif "-xlsx" in userflags:
                if "-y" in userflags:
                    return fio.readxlsx(filetoread, "-y", comparams)
                elif "-xy" in userflags:
                    return fio.readxlsx(filetoread, "-xy", comparams)
                elif "-xyz" in userflags:
                    return fio.readxlsx(filetoread, "-xyz", comparams)
                elif "-ye" in userflags:
                    return fio.readxlsx(filetoread, "-ye", comparams)
                elif "-xye" in userflags:
                    return fio.readxlsx(filetoread, "-xye", comparams)
                elif "-xeye" in userflags:
                    return fio.readxlsx(filetoread, "-xeye", comparams)
                elif "-xeyeze" in userflags:
                    return fio.readxlsx(filetoread, "-xeyeze", comparams)
                elif "-cye" in userflags:
                    return fio.readtxt(filetoread, "-cye", comparams)
                elif "-cyz" in userflags:
                    return fio.readtxt(filetoread, "-cyz", comparams)
                elif "-ccz" in userflags:
                    return fio.readtxt(filetoread, "-ccz", comparams)
                else:
                    return "No xlsx file structure given."
            elif "-svb" in userflags:
                if "-y" in userflags:
                    return fio.readsvb(filetoread, "-y", comparams)
                elif "-xy" in userflags:
                    return fio.readsvb(filetoread, "-xy", comparams)
                elif "-xye" in userflags:
                    return fio.readsvb(filetoread, "-xye", comparams)
                else:
                    return "No svb file structure given."
            elif "-fb" in userflags:
                return fio.readfb(filetoread)
            elif "-pvk" in userflags:
                return fio.readpvk(filetoread)
            elif "-i3x" in userflags:
                return fio.readi3x(filetoread)
            else:
                return "Invalid file type or structure given!"
        except:
            return "Specified File Format Does Not Match File Structure"

    def do_wri(self, *args):
        """\nCommand: WRIte\n
        Description: Write matrix data to various file types.

        Example Usage:
        \twri -txt  c:\data\data.txt (Wrtie matrix to tab delimited txt)
        \twri -xlsx c:\data\data.txt (Write matrix to xlsx file with figures)

        Default Input: N/A

        Default Options: -txt -xy  (if no file format given and extension matches a format, the xlsx extension will be used)

        Usage Options:
        \tN/A

        File Format Options:
        \t-txt (ascii text, tab or comma delimited)
        \t-xlsx (excel xlsx)

        """
        inparse = inputprocessing.InputParser()
        if not inparse(args):
            return "Invalid input!"
        formatoptions = ["txt", "xlsx"]
        userflags = inparse.cmdflags
        comparams = inparse.userinput

        if len(inparse.modifiers) == 0:
            return "No output file specified!"
        else:
            filetowrite = inparse.modifiers[0]
            if os.path.exists(filetowrite):
                return "Filename already exists!"
        try:
            ext = filetowrite.split('.')[-1]
            if ext.lower() == 'csv':
                ext = 'txt'
            if ext in formatoptions and ('-' + ext) not in userflags:
                userflags.append('-' + ext)
        except:
            ext = None
        if len(userflags) == 0:
            userflags.append("-xlsx")
        try:
            fio = fileio.IO(self.inst)
            if "-txt" in userflags:
                    return "No txt file structure programmed!"
            elif "-xlsx" in userflags:
                fio.writexlsx(filetowrite, sheet_name='Output', header_list=[], col_width_list=[], row_heights=152)
            else:
                return "Invalid file type or structure given!"
        except:
            e = sys.exc_info()[0]
            return "<p>Error: %s</p>" % e
        return "Data Written Successfully!"

    def do_test(self, *args):
        """"TESTING"""
        return "Testing"

    def do_exec(self, *args):
        """\nCommand: EXECute\n
        Description: Runs script of native commands from a txt file, line-by-line.

        Example Usage:
        \texec c:\data\script.txt (Executes line-by-line Pyvuka commands from specified file)
        \texec -script script_name (Executes line-by-line Pyvuka commands from specified script located in pyvuka Script directory)
        \texec -i c:\data\script.txt (Executes line-by-line Pyvuka commands from specified file, in interactive mode)
        \texec -module filename.py def_function (Executes custom Python function from module python files within Pyvuka Modules directory)

        Default Input: N/A

        Default Options: N/A

        Options:
        \t-i (interactive mode)
        """

        inparse = inputprocessing.InputParser()
        if not inparse(args):
            return "Invalid input!"
        userflags = inparse.cmdflags
        comparams = inparse.userinput

        if len(inparse.modifiers) == 0:
            return "No input file specified!"
        else:
            if len(inparse.modifiers) > 2 and "-module" in userflags:
                fixmethod = ' '.join(inparse.modifiers[1:])
                inparse.modifiers = [inparse.modifiers[0], fixmethod]
            filetoread = inparse.modifiers[0]
            interactivemode = False
            if "-module" in userflags:
                runmodule = "Modules."
                inparse.modifiers = [x.strip() for x in inparse.modifiers]
                for val in inparse.modifiers:
                    if val[-3:].lower() == ".py":
                        runmodule += val[:-2]
                        break
                for val in inparse.modifiers:
                    if ".py" not in val.lower():
                        runmodule += val
                        break
                if runmodule[-1:] != ")":
                    runmodule += "()"
                if runmodule[-2:] == '()':
                    runmodule = str(runmodule[:-2] + '(self.inst)')
                exec(runmodule)
                return "\nModule Run Successfully! Command Entered: " + runmodule
            if "-i" in userflags:
                interactivemode = True
            if "-script" in userflags:
                if '.' not in filetoread:
                    filetoread += ".txt"
                filetoread = os.path.join(os.path.dirname(os.path.abspath(__file__)), "Scripts", filetoread)
                if not os.path.exists(filetoread):
                    return "File Does Not Exist!"
                print("\nRunning Script: [" + filetoread + "]\tInteractive Mode: " + str(interactivemode) + "\n")
                return self.runscript(filetoread, interactivemode)
        return "Script Ran Successfully!"

    def do_cwd(self, *args):
        """\nCommand: Current Working Directory\n
        Description: Displays the current working directory.

        Example Usage:
        \tcwd (Displays current working directory)
        \tcwd -set c:\data\ (Sets current working directory)
        \tcwd -default (Sets current working directory to Pyvuka install directory)
        \tcwd -output (Sets output directory for write related modules)

        Default Input: N/A

        Default Options: N/A

        Options:
        \t-set (changes current working directory)
        """
        args_list = list(args)
        inparse = inputprocessing.InputParser()
        userflags = inparse.cmdflags

        if "-default" in userflags:
            self.inst.data.directories.working.set(os.path.dirname(os.path.abspath(__file__)))
            return "Current working directory SET to: " + self.inst.data.directories.working.get()

        if not inparse.modifiers:
            # inparse.modifiers is populated when path is entered.  Spaces and numbers in path are removed in this
            # object.  So we revert to args to determine path input for '-set'
            if '-set' in args_list:
                args_list.remove('-set')
                arg_dir = os.path.abspath(" ".join(args_list))
                if os.path.isdir(arg_dir):
                    self.inst.data.directories.working.set(arg_dir)
                    return f"Current working directory SET to: {self.inst.data.directories.working.get()}"
                else:
                    return f"Invalid Directory! Current working directory: {self.inst.data.directories.working.get()}"
            elif '-output' in args_list:
                args_list.remove('-output')
                arg_dir = " ".join(args_list)
                path_check = [os.path.exists(arg_dir), os.access(os.path.dirname(arg_dir), os.W_OK)]
                if True not in path_check:
                        return "Invalid Directory! No output directory has been set!"
                self.inst.data.directories.output.set(arg_dir)
                return f"Current output directory SET to: {self.inst.data.directories.output.get()}"
        return f"No changes made. Current working directory: {self.inst.data.directories.working.get()}"

    def do_cl(self, *args):
        """\nCommand: CLear\n
        Description: Clears buffer data by buffer range

        Example Usage:
        \tcl 1       (Clears first buffer)
        \tcl 1 10    (Clears buffers 1-10, inclusive)
        \tcl 1 10 2  (Clears buffers 1-10, every 2. Removes buffers: 1,3,5,7,9)
        \tcl all     (Clears all buffers)

        Default Inputs: N/A

        Default Options: N/A

        Options: N/A"""
        inparse = inputprocessing.InputParser()
        lastbuffer = self.inst.data.matrix.length()

        inparse.prompt = ["First Buffer", "Last Buffer", "Index Increment"]
        inparse.inputbounds = [[1, lastbuffer], [1, lastbuffer], [1, lastbuffer]]
        inparse.defaultinput = ['1', lastbuffer, '1']
        if not inparse(args):
            return "Invalid input!"
        else:
            if len(inparse.userinput) == 1:
                inparse.userinput.append(inparse.userinput[0])
                inparse.userinput.append('1')
            elif len(inparse.userinput) == 2:
                inparse.userinput.append('1')
            if "all" in inparse.modifiers:
                self.inst.data.matrix.clear()
                return "All Data Cleared!"
            gotparams = inparse.getparams()
        if not gotparams:
            return "No Buffers Were Cleared!"
        else:
            comparams = [int(val) for val in inparse.userinput]
            delarray = []
            for i in range(comparams[0], comparams[1]+1, comparams[2]):
                delarray.append(i-1)
            delarray.sort()
            for i in list(reversed(delarray)):
                self.inst.data.matrix.remove_buffer_by_number(i+1)
            delarray = [val+1 for val in delarray]
            return "Buffers: " + str(delarray) + " removed!"

    def do_now(self, *args):
        """\nCommand: NOW\n
        Description: Displays current buffer information as it exists 'NOW'

        Example Usage: now

        Default Input: N/A

        Default Options: N/A

        Options: N/A"""
        if self.inst.data.matrix.length() == 0:
            return "No data in memory!"
        else:
            # header
            print('{:7s}\t{:8s}\t{:18s}\t{:18s}\t{:18s}\t{:50s}'.format("Buffer#", "#Points", "   X    ", "   Y     ",
                      "   Z     ", 21*' ' + "Comments" + 21*' '))
            # boarder
            print('{:7s}\t{:8s}\t{:18s}\t{:18s}\t{:18s}\t{:50s}\n'.format(7*'-', 8*'-', 18*'-', 18*'-', 18*'-', 50*'-'))
            boarder = [7*'-', 8*'-', 18*'-', 18*'-', 18*'-', 50*'-']

            # buffer info
            for i in range(1, self.inst.data.matrix.length()+1, 1):
                buffer = self.inst.data.matrix.buffer(i)
                pnt = ' '
                x = ' '
                y = ' '
                z = ' '
                comments = ' '

                # get point #
                pnt = int(max([buffer.data.x.length(), buffer.category.x.length(),
                               buffer.data.y.length(), buffer.category.y.length(),
                               buffer.data.z.length(), buffer.category.z.length()]))
                if pnt < 1:
                    pnt = '????'

                # Get X range
                if buffer.data.x.length() > 0:
                    x = "{:.2f}".format(buffer.data.x.min()) + ' - ' + "{:.2f}".format(buffer.data.x.max())
                elif buffer.category.x.length() > 0:
                    xi = buffer.category.x.value_at_index(0)
                    xf = buffer.category.x.value_at_index(-1)
                    if len(xi) > 4:
                        xi = xi[:4] + "..."
                    if len(xf) > 4:
                        xf = xf[:4] + "..."
                    x = xi + ' - ' + xf

                # Get Y range
                if buffer.data.y.length() > 0:
                    y = "{:.2f}".format(buffer.data.y.min()) + ' - ' + "{:.2f}".format(buffer.data.y.max())
                elif buffer.category.y.length() > 0:
                    yi = buffer.category.y.value_at_index(0)
                    yf = buffer.category.y.value_at_index(-1)
                    if len(yi) > 4:
                        yi = yi[:4] + "..."
                    if len(yf) > 4:
                        yf = yf[:4] + "..."
                    y = yi + ' - ' + yf

                # Get Z range
                if buffer.data.z.length() > 0:
                    z = "{:.2f}".format(buffer.data.z.min()) + ' - ' + "{:.2f}".format(buffer.data.z.max())
                elif buffer.category.z.length() > 0:
                    zi = buffer.category.z.value_at_index(0)
                    zf = buffer.category.z.value_at_index(-1)
                    if len(zi) > 4:
                        yi = yi[:4] + "..."
                    if len(zf) > 4:
                        zf = zf[:4] + "..."
                    y = zi + ' - ' + zf

                # Get comments
                comments = buffer.comments.all_as_string()
                if len(comments) > 50:
                    comments = comments[:-3] + "..."

                print('{:7s}\t{:8s}\t{:18s}\t{:18s}\t{:18s}\t{:50s}'.format(str(i), str(pnt), x, y, z, comments))

        return '\n'

    def do_sbf(self, *args):
        """\nCommand: Subtract Buffer From\n
        Description:
        \tSubtracts indicated buffer from range of buffers (y-values), point-wise. Existing buffers are modified.
        \tSubtraction automatically truncates the buffers to the smallest size

        Example Usage:
        \t sbf 1 1 10 (subtracts data in buffer 1 from all buffers in range 1-10)
        \t sbf 1 10   (subtracts data in buffer 1 from only buffer 10)

        Default Input: N/A

        Default Options: N/A

        Options: N/A
        """
        inparse = inputprocessing.InputParser()
        lastbuffer = self.inst.data.matrix.length()
        inparse.prompt = ["Buffer to Subtract", "First Buffer", "Last Buffer"]
        inparse.inputbounds = [[1, lastbuffer], [1, lastbuffer], [1, lastbuffer]]
        inparse.defaultinput = ['1', '1', str(lastbuffer)]
        if not inparse(args):
            return "Invalid input!"
        else:
            if len(inparse.userinput) == 2:
                inparse.userinput.append(inparse.userinput[2])
            gotparams = inparse.getparams()
        if not gotparams:
            return "No buffers to subtract!"
        else:
            comparams = [int(val) for val in inparse.userinput]
            subtractionbuffer = self.inst.data.matrix.buffer(comparams[0]).data.y.get()
            subtractionye = self.inst.data.matrix.buffer(comparams[0]).data.ye.get()
            for i in range(comparams[1], comparams[2]+1, 1):
                buffer = self.inst.data.matrix.buffer(i)
                data_y = buffer.data.y.get()
                error_y = buffer.data.ye.get()
                minlen = min([len(data_y), len(subtractionbuffer)])
                data_y = data_y[:minlen] - subtractionbuffer[:minlen]
                if len(error_y) > 0 and len(subtractionye) > 0:
                    error_y = math.pow(math.pow(error_y[:minlen], 2) + math.pow(subtractionye[:minlen], 2), 0.5)
                buffer.comments.add("Buffer[" + str(comparams[0]) + "] Subtracted Data")

                buffer.data.y.set(data_y)
                buffer.data.ye.set(error_y)
                if buffer.data.x.length() > 0:
                    buffer.data.x.set(buffer.data.x.get()[:minlen])
                if buffer.data.xe.length() > 0:
                    buffer.data.xe.set(buffer.data.xe.get()[:minlen])
                if buffer.data.z.length() > 0:
                    buffer.data.z.set(buffer.data.z.get()[:minlen])
                if buffer.data.ze.length() > 0:
                    buffer.data.ze.set(buffer.data.ze.get()[:minlen])
            return "Buffer Subtraction Completed!"

    def do_res(self, *args):
        """\nCommand: RESample Data\n
        Description:
        \tResamples data within a buffer range to the number of points indicated bvy the used.
        \tResampled data is appended to the end of the data matrix as copied buffers.

        Example Usage:
        \tres 1 1 100   (resamples data in buffer 1 to have 100 data points)
        \tsbf 1 10 100  (resamples data in buffers 1 through 10 to have 100 data points)

        Default Input: N/A

        Default Options: N/A

        Options: NOT YET ADDED
        \t-log         (logarithmically resamples data)
        \t-antilog     (anti-logarithmically resamples data)
        \t-rand        (randomly resamples data)

        Notes:
        \tMinimum number of points to result from resampling is 3.
        """
        inparse = inputprocessing.InputParser()
        flags = inparse.cmdflags
        lastbuffer = self.inst.data.matrix.length()

        shortest = self.inst.data.matrix.shortest_x_length()

        inparse.prompt = ["First Buffer", "Last Buffer", "Number of Resulting Points"]
        inparse.inputbounds = [[1, lastbuffer], [1, lastbuffer], [3, shortest]]
        inparse.defaultinput = ['1', str(lastbuffer), str(shortest)]
        if not inparse(args):
            return "Invalid Input!"
        if not inparse.getparams():
            return"\nNo Data Was Resampled!"
        inparse.userinput = [int(val) for val in inparse.userinput]
        firstbuffer, lastbuffer, npoints = inparse.userinput

        for i in range(firstbuffer, lastbuffer+1):
            buffer = copy.deepcopy(self.inst.data.matrix.buffer(i))
            xtemp = []
            xetemp = []
            ytemp = []
            yetemp = []
            ztemp = []
            zetemp = []
            xlen = buffer.data.x.length()
            if xlen <= npoints:
                self.inst.data.matrix.add_buffer(buffer)
                continue
            stepsize = int(np.floor(xlen/npoints))
            if stepsize > 1:
                for j in range(0, xlen, stepsize):
                    if buffer.data.x.length() > 0:
                        xtemp.append(buffer.data.x.value_at_index(j))
                    if buffer.data.xe.length() > 0:
                        xetemp.append(buffer.data.xe.value_at_index(j))
                    if buffer.data.y.length() > 0:
                        ytemp.append(buffer.data.y.value_at_index(j))
                    if buffer.data.ye.length() > 0:
                        yetemp.append(buffer.data.ye.value_at_index(j))
                    if buffer.data.z.length() > 0:
                        ztemp.append(buffer.data.z.value_at_index(j))
                    if buffer.data.ze.length() > 0:
                        zetemp.append(buffer.data.ze.value_at_index(j))
            else:
                xtemp = buffer.data.x.get()
                xetemp = buffer.data.xe.get()
                ytemp = buffer.data.y.get()
                yetemp = buffer.data.ye.get()
                ztemp = buffer.data.z.get()
                zetemp = buffer.data.ze.get()
            if len(xtemp) > npoints and npoints > 3:
                if len(xtemp) > 0:
                    temp = xtemp[1:-1][:npoints-2]
                    temp.append(xtemp[-1])
                    xtemp = xtemp[:1] + temp
                if len(xetemp) > 0:
                    temp = xetemp[1:-1][:npoints-2]
                    temp.append(xetemp[-1])
                    xetemp = xetemp[:1] + temp
                if len(ytemp) > 0:
                    temp = ytemp[1:-1][:npoints-2]
                    temp.append(ytemp[-1])
                    ytemp = ytemp[:1] + temp
                if len(yetemp) > 0:
                    temp = yetemp[1:-1][:npoints-2]
                    temp.append(yetemp[-1])
                    yetemp = yetemp[:1] + temp
                if len(ztemp) > 0:
                    temp = ztemp[1:-1][:npoints-2]
                    temp.append(ztemp[-1])
                    ztemp = ztemp[:1] + temp
                if len(zetemp) > 0:
                    temp = zetemp[1:-1][:npoints-2]
                    temp.append(zetemp[-1])
                    zetemp = zetemp[:1] + temp
            buffer.data.x.set(xtemp)
            buffer.data.xe.set(xetemp)
            buffer.data.y.set(ytemp)
            buffer.data.ye.set(yetemp)
            buffer.data.z.set(ztemp)
            buffer.data.ze.set(zetemp)
            buffer.comments.add("Data Resampled to %s points" % npoints)
            self.inst.data.matrix.add_buffer(buffer)
        return '\nData Successfully Resampled!'

    def do_tri(self, *args):
        """\nCommand: TRIm Data\n
        Description:
        \tTrims data to index or X-value range defined by the user.
        \tTrimmed data overwrites the existing buffer.

        Example Usage:
        \ttri 1 2 10 1000     (trims data in buffers 1 through 2 to only include the points indexed from 10 to 1000)
        \ttri 1 2 0 100 -val  (trims data in buffers 1 through 2 to only include the points with X values between 0 and 100, inclusive)

        Default Input: N/A

        Default Options: N/A

        Options: NOT YET ADDED
        \t-v       (trims data based on X-val instead of point index)
        \t-val     (alias for above)
        \t-value   (alias for above)

        Notes:
        \tPoint index begins at 1, not 0
        """
        inparse = inputprocessing.InputParser()
        lastbuffer = self.inst.data.matrix.length()
        shortest = self.inst.data.matrix.shortest_x_length()
        useval = False

        inparse.prompt = ["First Buffer", "Last Buffer", "First Point Index", "Last Point Index"]
        inparse.inputbounds = [[1, lastbuffer], [1, lastbuffer], [1, shortest], [1, shortest]]
        inparse.defaultinput = ['1', lastbuffer, '1', str(shortest)]
        if '-v' in args or '-val' in args or '-value' in args:
            useval = True
            inparse.prompt = ["First Buffer", "Last Buffer", "Min X-val", "Max X-val"]
            inparse.inputbounds = [[1, lastbuffer], [1, lastbuffer], [-np.inf, np.inf], [-np.inf, np.inf]]
            inparse.defaultinput = ['1', lastbuffer,
                                    str(self.inst.data.matrix.buffer(1).data.x.get()[0]),
                                    str(self.inst.data.matrix.buffer(1).data.x.get()[-1])]
        if not inparse(args):
            return "Invalid Input!"
        if not inparse.getparams():
            return"\nNo Data Was Trimmed!"

        firstbuffer, lastbuffer, indexi, indexf = [int(float(x)) for x in inparse.userinput]
        for i in range(firstbuffer, lastbuffer + 1):
            buffer = self.inst.data.matrix.buffer(i)

            if useval:
                firstbuffer, lastbuffer, vali, valf = inparse.userinput
                indexi = buffer.data.x.nearest_index_to_value(vali)
                indexf = buffer.data.x.nearest_index_to_value(valf)

            low = min([indexi, indexf])
            high = max([indexi, indexf])
            if buffer.data.x.length() > 0:
                buffer.data.x.set(buffer.data.x.get()[low:high])
            if buffer.data.xe.length() > 0:
                buffer.data.xe.set(buffer.data.xe.get()[low:high])
            if buffer.data.y.length() > 0:
                buffer.data.y.set(buffer.data.y.get()[low:high])
            if buffer.data.ye.length() > 0:
                buffer.data.ye.set(buffer.data.ye.get()[low:high])
            if buffer.data.z.length() > 0:
                buffer.data.z.set(buffer.data.z.get()[low:high])
            if buffer.data.ze.length() > 0:
                buffer.data.ze.set(buffer.data.ze.get()[low:high])

            self.inst.data.matrix.set_buffer_by_number(buffer, i)
        return '\nData Successfully Trimmed!'

    def do_shi(self, *args):
        """\nCommand: SHIft Data\n
        Description:
        \tShifts data to by adding value defined by the user along the user defined axis. Default axis is X.
        \tShifted data overwrites the existing buffer.

        Example Usage:
        \tshi 1 2 -100        (Adds -100 to each X value in buffers 1 through 2)
        \tshi -y 1 2 -100     (Adds -100 to each Y value in buffers 1 through 2)

        Default Input: N/A

        Default Options: -x

        Options: NOT YET ADDED
        \t-x     (adds defined constant all X-value)
        \t-y     (adds defined constant all Y-value)
        \t-z     (adds defined constant all Z-value)

        Notes:
        \tN/A
        """
        inparse = inputprocessing.InputParser()
        lastbuffer = self.inst.data.matrix.length()

        inparse.prompt = ["First Buffer", "Last Buffer", "Constant Value"]
        inparse.inputbounds = [[1, lastbuffer], [1, lastbuffer], [-np.inf, np.inf]]
        inparse.defaultinput = [1, lastbuffer, 0.00]

        if not inparse(args):
            return "Invalid Input!"
        if not inparse.getparams():
            return"\nNo Data Was Shifted!"

        inparse.userinput = [int(val) for val in inparse.userinput]
        firstbuffer, lastbuffer, const = inparse.userinput

        for i in range(firstbuffer, lastbuffer + 1):
            buffer = self.inst.data.matrix.buffer(i)
            # add const to Z axis
            if "-z" in inparse.cmdflags:
                buffer.data.z.set(buffer.data.z.get() + const)
                buffer.plot.axis.z.lines.set(buffer.plot.axis.z.lines.get() + const)
            # add const to Y axis
            if "-y" in inparse.cmdflags:
                buffer.data.y.set(buffer.data.y.get() + const)
                buffer.plot.axis.y.lines.set(buffer.plot.axis.y.lines.get() + const)
            # add const to X axis if specified or as default with no flags
            if ("-x" in inparse.cmdflags) or ("-y" not in inparse.cmdflags and "-z" not in inparse.cmdflags):
                buffer.data.x.set(buffer.data.x.get() + const)
                buffer.plot.axis.x.lines.set(buffer.plot.axis.x.lines.get() + const)
        return '\nData Successfully Shifted!'

    def do_mul(self, *args):
        """\nCommand: MULtiply data by constant\n
        Description:
        \tSkews data by multiplying a constant value, defined by the user, along the user defined axis. Default axis is X.
        \tMultiplied data overwrites the existing buffer.

        Example Usage:
        \tmul 1 2 0.5        (Multiplies each Y value by 0.5 in buffers 1 through 2)
        \tmul -x 1 2 0.5     (Multiplies each X value by 0.5 in buffers 1 through 2)

        Default Input: N/A

        Default Options: -y

        \t-x     (multiplies defined constant all X-value)
        \t-y     (multiplies defined constant all Y-value)
        \t-z     (multiplies defined constant all Z-value)

        Notes:
        \tN/A
        """
        inparse = inputprocessing.InputParser()
        lastbuffer = self.inst.data.matrix.length()

        inparse.prompt = ["First Buffer", "Last Buffer", "Constant Value"]
        inparse.inputbounds = [[1, lastbuffer], [1, lastbuffer], [-np.inf, np.inf]]
        inparse.defaultinput = [1, lastbuffer, 0.00]

        if not inparse(args):
            return "Invalid Input!"
        if not inparse.getparams():
            return "\nNo Data Was Multiplied!"

        inparse.userinput = [int(val) for val in inparse.userinput]
        firstbuffer, lastbuffer, const = inparse.userinput

        for i in range(firstbuffer, lastbuffer + 1):
            buffer = self.inst.data.matrix.buffer(i)
            # mul const to Z axis
            if "-z" in inparse.cmdflags:
                buffer.data.z.set(buffer.data.z.get() * const)
            # mul const to Y axis
            if "-x" in inparse.cmdflags:
                buffer.data.x.set(buffer.data.x.get() * const)
            # mul const to X axis if specified or as default with no flags
            if ("-y" in inparse.cmdflags) or ("-x" not in inparse.cmdflags and "-z" not in inparse.cmdflags):
                buffer.data.y.set(buffer.data.y.get() * const)
        return '\nData Successfully Multiplied!'

    def do_mod(self, *args):
        """\nCommand: MODel data with current parameters\n
        Description:
        \tModels defined function with defined parameters

        Example Usage:
        \tmod 1 2         (Models function for buffers 1 through 2)
        \tmod -all        (Models function for all buffers)

        Default Input: N/A

        Default Options: N/A

        \t-all     (Models function for all buffers)

        Notes:
        \tN/A
        """
        inparse = inputprocessing.InputParser()
        lastbuffer = self.inst.data.matrix.length()
        firstbuffer = 1

        if '-all' not in args and 'all' not in args:
            inparse.prompt = ["First Buffer", "Last Buffer"]
            inparse.inputbounds = [[firstbuffer, lastbuffer], [firstbuffer, lastbuffer]]
            inparse.defaultinput = [firstbuffer, lastbuffer]

            if not inparse(args):
                return "Invalid Input!"
            if not inparse.getparams():
                return "\nNo Data Was Modeled!"

            inparse.userinput = [int(val) for val in inparse.userinput]
            firstbuffer, lastbuffer= inparse.userinput

        xr = self.inst.data.plot_limits.x_range.get()
        yr = self.inst.data.plot_limits.y_range.get()
        zr = self.inst.data.plot_limits.z_range.get()
        br = self.inst.data.plot_limits.buffer_range.get()
        ma = self.inst.data.plot_limits.is_active

        pti = 1 if not xr else min(xr)
        ptf = self.inst.data.matrix.buffer(1).data.x.length() if not xr else max(xr)
        self(f'fix all')
        self(f'pl off')
        self(f'pl {firstbuffer} {lastbuffer} {pti} {ptf}')
        self(f'pl on')
        self(f'fit')
        self(f'pl off')
        self(f'fre all')

        self.inst.data.plot_limits.x_range.set(xr)
        self.inst.data.plot_limits.y_range.set(yr)
        self.inst.data.plot_limits.z_range.set(zr)
        self.inst.data.plot_limits.buffer_range.set(br)
        self.inst.data.plot_limits.is_active = ma
        return '\nData Successfully Modeled!'

    def do_ori(self, *args):
        """\nCommand: ORIgin\n
        Description:
        \tSets desired point index to origin (0,0,0).
        \tModified data overwrites the existing buffer.

        Example Usage:
        \tori 1 2 10        (Sets point #10 in buffers 1 through 2 as the XYZ origin)
        \tori -y 1 2 10     (Sets point #10 in buffers 1 through 2 as Y=0)

        Default Input: N/A

        Default Options: -x -y -z

        Options: NOT YET ADDED
        \t-x     (X=0 for defined point index)
        \t-y     (Y=0 for defined point index)
        \t-z     (Z=0 for defined point index)

        Notes:
        \tN/A
        """
        inparse = inputprocessing.InputParser()
        lastbuffer = self.inst.data.matrix.length()

        inparse.prompt = ["First Buffer", "Last Buffer", "Point Index"]
        inparse.inputbounds = [[1, lastbuffer], [1, lastbuffer], [1, np.inf]]
        inparse.defaultinput = [1, lastbuffer, 1]

        if not inparse(args):
            return "Invalid Input!"
        if not inparse.getparams():
            return "\nNo Data Was Shifteded!"

        inparse.userinput = [int(val) for val in inparse.userinput]
        firstbuffer, lastbuffer, point_index = inparse.userinput
        point_index = point_index - 1  # 1->inf in user input, 0->inf in array

        if not inparse.cmdflags:
            inparse.cmdflags = ['-x', '-y', '-z']

        for i in range(firstbuffer, lastbuffer + 1):
            buffer = self.inst.data.matrix.buffer(i)
            # sub const to Z axis
            if "-z" in inparse.cmdflags and buffer.data.z.length() > 0:
                const = buffer.data.z.value_at_index(point_index)
                buffer.data.z.set(buffer.data.z.get() - const)
                buffer.plot.axis.z.lines.set(buffer.plot.axis.z.lines.get() - const)
            # sub const to Y axis
            if "-y" in inparse.cmdflags:
                const = buffer.data.y.value_at_index(point_index)
                buffer.data.y.set(buffer.data.y.get() - const)
                buffer.plot.axis.y.lines.set(buffer.plot.axis.y.lines.get() - const)
            # sub const to X axis
            if "-x" in inparse.cmdflags:
                const = buffer.data.x.value_at_index(point_index)
                buffer.data.x.set(buffer.data.x.get() - const)
                buffer.plot.axis.x.lines.set(buffer.plot.axis.x.lines.get() - const)
        return '\nData Successfully Shifted to Origin!'

    def do_dr(self, *args):
        """\nCommand: DRaw\n
        Description: Draws requested plots in graphical output.

        Example Usage:
        \tdr (draws first buffer)
        \tdr all (draws all buffers)
        \tdr 1 10 2 (draws buffers 1-10, every 2.  Output is drawing 1, 3, 5, 7, 9)
        \tdr -l 1 3 5 7 9  (Draws a list of buffers. Output is drawing 1, 3, 5, 7, 9)

        Default Input: 1 1 1

        Default Options: N/A

        Options:
        \t -l (reads user input as list instead of incremental plotting)

        NOTES: Legacy Savuka "stack depth" (sd) is now controlled at the individual buffer level.  Similarly the
        following legacy options are depricated and are controlled at the individual buffer level:
        \t-err
        \t-nolegend
        \t-noweight
        Multiplot parameters will override individual settings."""
        inparse = inputprocessing.InputParser()
        if self.inst.data.plot_limits.is_active:
            firstbuffer, lastbuffer = self.inst.data.plot_limits.buffer_range.get()
        else:
            lastbuffer = self.inst.data.matrix.length()
            firstbuffer = 1
        inparse.prompt = ["First Buffer", "Last Buffer", "Index Increment"]
        inparse.inputbounds = [[firstbuffer, lastbuffer], [firstbuffer, lastbuffer], [firstbuffer, lastbuffer]]
        inparse.defaultinput = ['1', str(lastbuffer), '1']
        if not inparse(args):
            return "Invalid input!"
        modifiers = inparse.modifiers
        userflags = inparse.cmdflags
        comparams = inparse.userinput
        pltarray = []
        modifiers = [val.lower() for val in modifiers]
        for i in range(len(comparams)):
            if self.inst.data.plot_limits.is_active:
                comparams[i] = int(comparams[i]) if firstbuffer <= comparams[i] else firstbuffer
                comparams[i] = int(comparams[i]) if comparams[i] <= lastbuffer else lastbuffer
            else:
                comparams[i] = int(comparams[i]) if comparams[i] <= lastbuffer else lastbuffer

        plotwindow = plot.plotter(self.inst.data)
        if 'all' in modifiers:
            pltarray = [buf for buf in range(firstbuffer, lastbuffer + 1)]
        elif "-l" in userflags:
            comparams = [val for val in comparams]
            pltarray = comparams
        elif len(comparams) == 0:
            pltarray = [1]
        elif len(comparams) == 1:
            pltarray = [int(comparams[0])]
        elif len(comparams) == 2:
            for i in range(comparams[0], comparams[1] + 1):
                pltarray.append(i)
        elif len(comparams) > 2:
            for i in range(comparams[0], comparams[1] + 1, comparams[2]):
                pltarray.append(i)
        return plotwindow(pltarray)

    def do_sca(self, *args):
        """\nCommand: SCAn\n
               Description: Draws specified plots one at a time. Pressing any key will draw the next plot int he series.

               Example Usage:
               \tsca (Draws all buffers, one at a time)
               \tsca 1 10 2 (Draws buffers 1-10, every 2.  Output is drawing 1, 3, 5, 7, 9, one at a time)

               Default Input: 1 [last buffer index] 1

               Default Options: N/A

               Options:
               \t-auto (Automatically iterates through plots with 0.5 second delay between plots)
               \t-auto=# (Auto function with delay of 0.5 sec overridden by specified # of sec.

               NOTES: Multiplot parameters are ignored, individual settings will be used."""
        delay = 0.5  # time in seconds
        auto = False
        inparse = inputprocessing.InputParser()
        if self.inst.data.plot_limits.is_active:
            firstbuffer, lastbuffer = self.inst.data.plot_limits.buffer_range.get()
        else:
            lastbuffer = self.inst.data.matrix.length()
            firstbuffer = 1
        inparse.prompt = ["First Buffer", "Last Buffer", "Index Increment"]
        inparse.inputbounds = [[firstbuffer, lastbuffer], [firstbuffer, lastbuffer], [1, lastbuffer]]
        inparse.defaultinput = ['1', str(lastbuffer), '0']
        if not inparse(args):
            return "Invalid input!"
        userflags = inparse.cmdflags
        comparams = inparse.userinput
        if len(comparams) == 0:
            inparse.userinput = inparse.defaultinput
        plotwindow = plot.plotter(self.inst.data)
        pltarray = []
        comparams = [int(val) for val in comparams]
        userflags = [st for st in userflags if "-auto" in st]
        if len(userflags) > 0:
            auto = True
            try:
                delay = float(userflags[0].split('=')[1])
            except:
                delay = 0.5
            pltarray = [buf for buf in range(firstbuffer-1, lastbuffer)]
            return plotwindow(pltarray, scan=True, auto=auto, delay=delay)
        gotparams = inparse.getparams()
        if not gotparams:
            return "\nNo Data to Scan!"
        comparams = [int(val) for val in inparse.userinput]
        for i in range(comparams[0], comparams[1]+1, comparams[2]):
                pltarray.append(i-1)
        return plotwindow(pltarray, scan=True, auto=auto, delay=delay)

    def do_fun(self, *args):
        """\nCommand: FUNction\n
        Description: Sets the functions or string of functions to be used for fitting data.
        Entering an index of 0 will stop concatenating functions.

        Example Usage:
        \tfun            (shows current list of available functions)
        \tfun info       (displays available function indexes for which info is available)
        \tfun info 27    (displays detailed information about function with index 27)
        \tfun 2 1 0      (sets the selected functions to 2 and 1)

        Default Input: N/A

        Default Options: N/A

        Options: N/A"""
        fail = "Invalid Input!"
        success = "Please input default parameters (see command: ap).\nFunction(s) set: "
        fitter = fitfxns.datafit(self.inst)
        count = 1
        userin = 100
        fxnlist = fitter.getfxnlist()
        fxnlist = [int(val) for val in fxnlist]
        args = [val.lower() for val in args]
        # Parse user arguments
        inparse = inputprocessing.InputParser()
        inparse(args)

        if len(args) == 0:  # If no arguments provided, return list of available functions
            return fitter.showfxntable()
        elif self.inst.data.matrix.length() == 0:  # If data matrix is invalid, return call
            return "No Data in Memory! Functions Cannot Be Applied!"
        elif 'info' in inparse.modifiers:  # If info in args, show function info
            return fitter(args)
        elif len(inparse.userinput) > 0 and inparse.userinput[-1] == 0: # If function request is closed, define fitter
            for num in inparse.userinput[:-1]:
                try:
                    temp = int(num)
                    if temp in fxnlist:
                        fitter.funcindex.append(temp)
                except ValueError:
                    return fail
            if len(fitter.funcindex) == 0:
                return fail
            else:
                if fitter.applyfxns():
                    return success + str(fitter.funcindex)
                else:
                    return"\nInvalid Functions Specified!"
        elif len(inparse.userinput) > 0 and inparse.userinput[-1] != 0: # If  function request is not closed, append function to fitter definition
            for val in inparse.userinput:
                try:
                    temp = int(val)
                    if temp in fxnlist:
                        fitter.funcindex.append(temp)
                        count += 1
                except ValueError:
                    return fail

        # prompt user to complete fxn definition
        print("\nEnter functions to concatenate, enter 0 to stop.\n")
        while userin > 0:
            userin = input("Enter Function #" + str(count)+" :  ")
            try:
                userin = int(userin)
            except ValueError:
                print(fail)
            if userin in fxnlist:
                count += 1
                fitter.funcindex.append(userin)
            elif userin != 0:
                print("Not a Valid Function!")
        # Apply fxn definitions to fitter object and return success/fail
        if fitter.applyfxns():
            return success + str(fitter.funcindex)
        else:
            "\nInvalid Functions Specified!"

    def do_fit(self, *args):
        args = [val.lower() for val in args]
        ### Pre-flight check for any parameter linking ###
        linked = False
        plon = False
        if self.inst.data.plot_limits.is_active:
            firstbuffer, lastbuffer = self.inst.data.plot_limits.buffer_range.get()
            firstpoint, lastpoint = self.inst.data.plot_limits.x_range.get()
            plon = True
        else:
            lastbuffer = self.inst.data.matrix.length()
            firstbuffer = 1
            firstpoint = 1
            lastpoint = self.inst.data.matrix.shortest_x_length()

        for i in range(firstbuffer, lastbuffer + 1):
            buffer = self.inst.data.matrix.buffer(i)
            linked = True if len([x for x in buffer.fit.link.get() if x is not None]) >= 1 else False
            if linked:
                break

        ### if independent fit override, all links will be copied and removed for fit--added back after ###
        alllinks = []
        if "-ind" in args:
            for i in range(firstbuffer, lastbuffer + 1):
                buffer = self.inst.data.matrix.buffer(i)
                alllinks.append(buffer.fit.link.get())
            self("unl -all")
        ### If data is not linked or override to independent fitting and restore link and plot limit states###
        if not linked:
            print(fitfxns.datafit(self.inst).dofit(*args))
            self("pl off")
            if plon:
                com_list = [firstbuffer, lastbuffer, firstpoint, lastpoint]
                command = "pl {} {} {} {}".format(*com_list)
                self(command)
                self("pl on")
            else:
                self("pl off")
        if "-ind" in args:  ### if -ind flag, restore original links ###
            for i in range(firstbuffer, lastbuffer + 1):
                buffer = self.inst.data.matrix.buffer(i)
                buffer.fit.link.set(alllinks[i - 1])
            return "\nIndependent Fitting Complete!"
        elif not linked:
            return "\nIndependent Fitting Complete!"
        else:  ### if data is linked we run global fitting ###
            print(fitfxns.datafit(self.inst).dofit(*args))
            return "\nGlobal Fitting Complete!"

    def do_ap(self, *args):
        """\nCommand: Alter Parameters\n
        Description: Prompts users to enter parameters for specified buffers.
        Enter 'q' to exit prompt cycle.

        Example Usage:
        \tap                      (Prompt user for all parameter information)
        \tap -all                 (Run a single cycle to collect parameters, applies to all buffers within plot limits)
        \tap all                  (Run a single cycle to collect parameters, applies to all buffers within plot limits)
        \tap 1 -1.56 2.7          (Sets parameters 1 & 2  to -1.56 and 2.7, for buffer 1)
        \tap -all -1.56 2.7       (Sets parameters 1 & 2  to -1.56 and 2.7, for all buffers within plot limits)
        \tap -cp                  (Prompts user for buffer to copy parameters from and buffer range to copy parameters to)
        \tap -cp 1 2 10           (Copies parameters from buffer 1 to all buffers in range 2-10)

        Default Input: N/A

        Default Options: N/A

        Options:
        \t-all        (Applies parameter input to all buffers within plot limits)
        \t-cp         (Copies parameters from specified buffer to all buffers in specified range)
        \t-cpp        (Same as above)
        \t-copy       (Same as above)

        Modifiers
        \tall         (Applies parameter input to all buffers within plot limits)

        Notes: Parameter links and fix states are preserved with copy functions
        """
        inparse = inputprocessing.InputParser()
        inparse(args)
        userincount = len(inparse.userinput)
        applyall = False
        if self.inst.data.plot_limits.is_active:
            firstbuffer, lastbuffer = self.inst.data.plot_limits.buffer_range.get()
        elif userincount == 1:
            firstbuffer = int(inparse.userinput[0])
            lastbuffer = int(inparse.userinput[0])
        else:
            lastbuffer = self.inst.data.matrix.length()
            firstbuffer = 1
        if len(self.inst.data.matrix.buffer(firstbuffer).fit.function.get()) == 0:
            self.updatefixlnk()
            return "\nNo Function Defined. Try command: fun ."
        if "-all" in inparse.cmdflags or "all" in inparse.modifiers:
            applyall = True
        if "-cp" in inparse.cmdflags or "-cpp" in inparse.cmdflags or "-copy" in inparse.cmdflags:
            inparse.prompt = ["Buffer to copy parameters from"]
            inparse.inputbounds = [[firstbuffer, lastbuffer]]
            inparse.defaultinput = [str(firstbuffer)]
            userinsave = inparse.userinput
            inparse.userinput = inparse.userinput[:1]
            if not inparse.getparams():
                return "\nNo Parameters Were Copied!"
            parambuffer = int(inparse.userinput[0])
            inparse.userinput = userinsave[1:]
            if applyall:
                inparse.userinput = [firstbuffer, lastbuffer]
            inparse.prompt = ["First Buffer to Apply Parameters to", "Last buffer to apply parameters to"]
            inparse.inputbounds = [[firstbuffer,lastbuffer], [firstbuffer,lastbuffer]]
            inparse.defaultinput = [str(firstbuffer), str(lastbuffer)]
            if not inparse.getparams():
                return "\nNo Parameters Were Copied!"
            inparse.userinput = [int(val) for val in inparse.userinput]
            paramstocopy = self.inst.data.matrix.buffer(parambuffer).fit.parameter.get()
            for i in range(inparse.userinput[0], inparse.userinput[1]+1):
                self.inst.data.matrix.buffer(i).fit.parameter.set(paramstocopy[:])
            self.updatefixlnk()
            return "\nParameters copied from Buffer[" + str(parambuffer) + "] to Buffers[" + \
                   str(inparse.userinput[0]) + " - " + str(inparse.userinput[1]) + "] !"

        i = firstbuffer
        while i <= lastbuffer:
            fitparams = fitfxns.datafit(self.inst)
            fitparams.update(self.inst.data.matrix.buffer(i).fit.function_index.get())
            if not applyall:
                inparse.prompt = ["Buffer Number"]
                inparse.inputbounds = [[firstbuffer, lastbuffer]]
                inparse.defaultinput = [str(i)]
            inparse.prompt.extend([str(val) for val in fitparams.paramid])
            inparse.inputbounds.extend(fitparams.parambounds)
            inparse.defaultinput.extend([str(val) for val in fitparams.paramdefaults])
            inparse.previousvals = [str(i)] + [str(val) for val in self.inst.data.matrix.buffer(i).fit.parameter.get()]
            if not inparse.getparams():
                return "Changes Made Before Aborting Operation Have Been Kept!"
            inparse.userinput = [float(val) for val in inparse.userinput]
            i = int(inparse.userinput[0])
            if not applyall:
                self.inst.data.matrix.buffer(i).fit.parameter.set(inparse.userinput[1:])
            if applyall:
                #need use extend method to prevent linking dictionary values by reference // replaced with deepcopy
                for j in range(firstbuffer, lastbuffer + 1):
                    self.inst.data.matrix.buffer(j).fit.parameter.set(copy.deepcopy(inparse.userinput))
                self.updatefixlnk()
                return "\nAll Initial Parameters Set!"
            if userincount > 0:
                self.updatefixlnk()
                return "\nInitial Parameters Set for Buffer[%s]!" % int(inparse.userinput[0])
            inparse.clear()
            fitparams.clear()
            i += 1
        self.updatefixlnk()
        return "\nInitial Parameters Set!"

    def do_lnk(self, *args):
        """\nCommand: LiNK parameters\n
        Description: Prompts users to enter parameter linkage information.
        Enter 'q' to exit prompt cycle.

        Example Usage:
        \tlnk                     (Prompt user for all parameter linkage information)
        \tlnk -all                (Links each parameter across all buffers)
        \tlnk all                 (Links each parameter across all buffers)
        \tlnk 1                   (Links parameter 1 for all buffers within plot limits)
        \tlnk 2 7 1 2             (Links Buffer 2, parameter 7 to Buffer 1, Parameter 2)

        Default Input: N/A

        Default Options: N/A

        Options:
        \t-all        (Applies parameter input to all buffers within plot limits)

        Modifiers"
        \tall         (Applies parameter input to all buffers within plot limits)
        """
        fitparams = fitfxns.datafit(self.inst)
        inparse = inputprocessing.InputParser()
        inparse(args)
        applyall = False
        if self.inst.data.plot_limits.is_active:
            firstbuffer, lastbuffer = self.inst.data.plot_limits.buffer_range.get()
        else:
            lastbuffer = self.inst.data.matrix.length()
            firstbuffer = 1
        if len(self.inst.data.matrix.buffer(firstbuffer).fit.parameter.get()) < 1:
            return "\nNo Initial Parameters Defined! Try command: ap ."
        if "-all" in inparse.cmdflags or "all" in inparse.modifiers:
            applyall = True
        if len(args) == 1:
            if applyall:
                for i in range(firstbuffer, lastbuffer - 1):
                    buffer = self.inst.data.matrix.buffer(i)
                    fitparams.update(buffer.fit.function_index.get())
                    links = copy.deepcopy(buffer.fit.link.get())
                    if i != firstbuffer:
                        for j in range(len(links)):
                            links[j] = f"{fitparams.paramid[j]}_{j+1}_{firstbuffer}"
                    self.inst.data.matrix.buffer(i).fit.link.set(links)
                return "\nAll Parameters Linked!"
            elif len(inparse.userinput) > 0:
                try:
                    for i in range(firstbuffer, lastbuffer + 1):
                        fitparams.update(self.inst.data.matrix.buffer(i).fit.function_index.get())
                        if i != firstbuffer:
                            links = copy.deepcopy(self.inst.data.matrix.buffer(i).fit.link.get())
                            links[int(inparse.userinput[0]) - 1] = fitparams.paramid[int(inparse.userinput[0])-1] + "_{}_{}".format(int(inparse.userinput[0]), firstbuffer)
                            self.inst.data.matrix.buffer(i).fit.link.set(links)
                except:
                    return "\nInvalid Parameter Specified!"
                return "\nParameter %s Linked Across All Buffers!" % str(int(inparse.userinput[0]))
        else:
            inparse.prompt = ["Buffer to link", "Parameter to link", "Buffer to link to", "Parameter to link to"]
            inparse.inputbounds = [[firstbuffer, lastbuffer], [1, np.inf], [firstbuffer, lastbuffer], [1, np.inf]]
            inparse.defaultinput = [firstbuffer, 1, lastbuffer, 1]
            if not inparse.getparams():
                return"\nNo Buffers Were Linked!"
            fitparams.update(self.inst.data.matrix.buffer(int(inparse.userinput[0])).fit.function_index.get())
            input_buffers = [int(inparse.userinput[0]), int(inparse.userinput[2])]
            input_params = [int(inparse.userinput[1]), int(inparse.userinput[3])]
            buffer_to_link = max(input_buffers)
            floating_buffer = min(input_buffers)
            floating_parameter = input_params[input_buffers.index(floating_buffer)]
            linked_parameter = input_params[input_buffers.index(buffer_to_link)]
            try:
                links = self.inst.data.matrix.buffer(buffer_to_link).fit.link.get()
                links[linked_parameter - 1] = fitparams.paramid[floating_parameter - 1] + "_{}_{}".format(floating_parameter, floating_buffer)
                self.inst.data.matrix.buffer(inparse.userinput[0]).fit.link.set(links)
                return f"\nParameter {linked_parameter} of Buffer {buffer_to_link} Linked to Parameter {floating_parameter} of Buffer {floating_buffer}"
            except:
                return "\nInvalid Parameter Specified!"
        return "\nInvalid Command!"

    def do_unl(self, *args):
        """\nCommand: UNLink parameters\n
        Description: Prompts users to enter parameter linkage information.
        Enter 'q' to exit prompt cycle.

        Example Usage:
        \tunl                     (Prompt user for all parameter linkage information)
        \tunl -all                (Unlinks each parameter across all buffers)
        \tunl all                 (Unlinks each parameter across all buffers)
        \tunl 1                   (Unlinks parameter 1 for all buffers within plot limits)
        \tunl 1 3                 (Unlinks parameter 3 of Buffer 1)

        Default Input: N/A

        Default Options: N/A

        Options:
        \t-all        (Applies parameter input to all buffers within plot limits)

        Modifiers"
        \tall         (Applies parameter input to all buffers within plot limits)
        """
        inparse = inputprocessing.InputParser()
        inparse(args)
        applyall = False
        if self.inst.data.plot_limits.is_active:
            firstbuffer, lastbuffer = self.inst.data.plot_limits.buffer_range.get()
        else:
            lastbuffer = self.inst.data.matrix.length()
            firstbuffer = 1
        if len(self.inst.data.matrix.buffer(firstbuffer).fit.parameter.get()) < 1:
            return "\nNo Initial Parameters Defined! Try command: ap ."
        if "-all" in inparse.cmdflags or "all" in inparse.modifiers:
            applyall = True
        if len(args) == 1:
            if applyall:
                for i in range(firstbuffer, lastbuffer+1):
                    unlinks = [None] * len(self.inst.data.matrix.buffer(i).fit.link.get())
                    self.inst.data.matrix.buffer(i).fit.link.set(unlinks)
                return "\nAll Parameters Unlinked!"
            elif len(inparse.userinput) > 0:
                try:
                    for i in range(firstbuffer, lastbuffer+1):
                        links = self.inst.data.matrix.buffer(i).fit.link.get()
                        links[int(inparse.userinput[0]) - 1] = None
                        self.inst.data.matrix.buffer(i).fit.link.set(links)
                except:
                    return "\nInvalid Parameter Specified!"
                return "\nParameter %s Unlinked Across All Buffers!" % str(int(inparse.userinput[0]))
        else:
            inparse.prompt = ["Buffer to modify", "Parameter to unlink"]
            inparse.inputbounds = [[firstbuffer, lastbuffer], [1, np.inf]]
            inparse.defaultinput = [firstbuffer, 1]
            if not inparse.getparams():
                return "\nNo Parameter Was Linked!"
            try:
                links = self.inst.data.matrix.buffer(int(inparse.userinput[0])).fit.link.get()
                links[int(inparse.userinput[1]) - 1] = None
                self.inst.data.matrix.buffer(int(inparse.userinput[0])).fit.link.set(links)
                return "\nParameter " + str(inparse.userinput[1]) + " of Buffer " + \
                       str(inparse.userinput[0]) + " Unlinked!"
            except:
                return "\nInvalid Parameter Specified!"

        return "\nInvalid Command!"

    def do_fix(self, *args):
        """\nCommand: FIX parameters\n
        Description: Prompts users to enter parameter free/fix information.
        Enter 'q' to exit prompt cycle.

        Example Usage:
        \tfix                     (Prompt user for all parameter Fix information)
        \tfix -all                (Fixes each parameter across all buffers)
        \tfix all                 (Fixes each parameter across all buffers)
        \tfix 3                   (Fixes parameter 3 for all buffers within plot limits)
        \tfix 1 3                 (Fixes parameter 3 of Buffer 1)

        Default Input: N/A

        Default Options: N/A

        Options:
        \t-all        (Applies parameter input to all buffers within plot limits)

        Modifiers:
        \tall         (Applies parameter input to all buffers within plot limits)
        """
        inparse = inputprocessing.InputParser()
        inparse(args)
        applyall = False
        if self.inst.data.plot_limits.is_active:
            firstbuffer, lastbuffer = self.inst.data.plot_limits.buffer_range.get()
        else:
            lastbuffer = self.inst.data.matrix.length()
            firstbuffer = 1
        if len(self.inst.data.matrix.buffer(firstbuffer).fit.parameter.get()) < 1:
            return "\nNo Initial Parameters Defined! Try command: ap ."
        if "-all" in inparse.cmdflags or "all" in inparse.modifiers:
            applyall = True
        if len(args) == 1:
            if applyall:
                for i in range(firstbuffer, lastbuffer+1):
                    new_free = [False] * len(self.inst.data.matrix.buffer(i).fit.free.get())
                    self.inst.data.matrix.buffer(i).fit.free.set(new_free)
                return "\nAll Parameters Fixed!"
            else:
                for i in range(firstbuffer, lastbuffer+1):
                    free = self.inst.data.matrix.buffer(i).fit.free.get()
                    free[int(inparse.userinput[0]) - 1] = False
                    self.inst.data.matrix.buffer(i).fit.free.set(free)
            return "\nParameter %s Fixed!" % str(int(inparse.userinput[0]))
        else:
            inparse.prompt = ["Buffer to modify", "Parameter to fix"]
            inparse.inputbounds = [[firstbuffer, lastbuffer], [1, np.inf]]
            inparse.defaultinput = [firstbuffer, 1]
            if not inparse.getparams():
                return "\nNo Parameter Was Fixed!"
            free = self.inst.data.matrix.buffer(int(inparse.userinput[0])).fit.free.get()
            free[int(inparse.userinput[1]) - 1] = False
            self.inst.data.matrix.buffer(int(inparse.userinput[0])).fit.free.set(free)
            return "Parameter " + str(int(inparse.userinput[1])) + " of buffer " + \
                   str(int(inparse.userinput[1])) + " Fixed!"

    def do_fre(self, *args):
        """\nCommand: FREe parameters\n
        Description: Prompts users to enter parameter free/fix information.
        Enter 'q' to exit prompt cycle.

        Example Usage:
        \tfre                     (Prompt user for all parameter Free information)
        \tfre -all                (Frees each parameter across all buffers)
        \tfre all                 (Frees each parameter across all buffers)
        \tfre 3                   (Frees parameter 3 for all buffers within plot limits)
        \tfre 1 3                 (Frees parameter 3 of Buffer 1)

        Default Input: N/A

        Default Options: N/A

        Options:
        \t-all        (Applies parameter input to all buffers within plot limits)

        Modifiers:
        \tall         (Applies parameter input to all buffers within plot limits)
        """
        inparse = inputprocessing.InputParser()
        inparse(args)
        applyall = False
        if self.inst.data.plot_limits.is_active:
            firstbuffer, lastbuffer = self.inst.data.plot_limits.buffer_range.get()
        else:
            lastbuffer = self.inst.data.matrix.length()
            firstbuffer = 1
        if len(self.inst.data.matrix.buffer(firstbuffer).fit.parameter.get()) < 1:
            return "\nNo Initial Parameters Defined! Try command: ap ."
        if "-all" in inparse.cmdflags or "all" in inparse.modifiers:
            applyall = True
        if len(args) == 1:
            if applyall:
                for i in range(firstbuffer, lastbuffer+1):
                    free = [True] * len(self.inst.data.matrix.buffer(i).fit.free.get())
                    self.inst.data.matrix.buffer(i).fit.free.set(free)
                return "\nAll Parameters Freed!"
            else:
                for i in range(firstbuffer, lastbuffer+1):
                    free = self.inst.data.matrix.buffer(i).fit.free.get()
                    free[int(inparse.userinput[0]) - 1] = True
                    self.inst.data.matrix.buffer(i).fit.free.set(free)
            return "\nParameter %s Freed!" % str(int(inparse.userinput[0]))
        else:
            inparse.prompt = ["Buffer to modify", "Parameter to free"]
            inparse.inputbounds = [[firstbuffer, lastbuffer], [1, np.inf]]
            inparse.defaultinput = [firstbuffer, 1]
            if inparse.getparams():
                free = self.inst.data.matrix.buffer(int(inparse.userinput[0])).fit.free.get()
                free[int(inparse.userinput[1]) - 1] = True
                self.inst.data.matrix.buffer(int(inparse.userinput[0])).fit.free.set(free)
                return "Parameter " + str(int(inparse.userinput[1])) + " of buffer " + \
                       str(int(inparse.userinput[1])) + " Freed!"
            else:
                return "Changes Made Prior to Aborting Have Been Kept!"

    def do_lp(self, *args):
        """\nCommand: List Parameters\n
                Description: List parameter information for the specified buffer or range of buffers.

                Example Usage:
                \tlp                     (Prompt user for buffers to show parameter information of)
                \tlp -all                (Shows parameter information for all buffers all buffers)
                \tlp all                 (Shows parameter information for all buffers all buffers)
                \tlp 1                   (Shows parameter information for all buffer 1)
                \tlp 1 3                 (Shows parameter information for all buffers 1 through 3)

                Default Input: N/A

                Default Options: N/A

                Options:
                \t-all        (Shows parameter information for all buffers within plot limits)

                Modifiers:
                \tall         (Shows parameter information for all buffers within plot limits)
                """
        inparse = inputprocessing.InputParser()
        inparse(args)
        fitparams = fitfxns.datafit(self.inst)
        if self.inst.data.plot_limits.is_active:
            firstbuffer, lastbuffer = self.inst.data.plot_limits.buffer_range.get()
        else:
            lastbuffer = self.inst.data.matrix.length()
            firstbuffer = 1
        if len(self.inst.data.matrix.buffer(firstbuffer).fit.parameter.get()) < 1:
            return "\nNo Initial Parameters Defined! Try command: ap ."
        if "-all" in inparse.cmdflags or "all" in inparse.modifiers:
            inparse.userinput = [str(firstbuffer), str(lastbuffer)]
        if len(inparse.userinput) == 1:
            inparse.userinput.append(inparse.userinput[0])
        inparse.prompt = ["First Buffer", "Last Buffer"]
        inparse.inputbounds = [[firstbuffer, lastbuffer], [firstbuffer, lastbuffer]]
        inparse.defaultinput = [firstbuffer, lastbuffer]
        if not inparse.getparams():
            return "\nNo Changes Have Been Made!"
        firstbuffer, lastbuffer = inparse.userinput
        firstbuffer = int(firstbuffer)
        lastbuffer = int(lastbuffer)
        header = ["Buffer", "Parameter", "Value", "Error", "Linked", "Fixed"]
        line = '-'*120
        spacer = ' '*20
        table = '\n\n'
        for i in range(firstbuffer, lastbuffer+1):
            fitparams.update(self.inst.data.matrix.buffer(i).fit.function_index.get())
            if len(self.inst.data.matrix.buffer(i).fit.parameter_error.get()) != len(self.inst.data.matrix.buffer(i).fit.parameter.get()):
                self.inst.data.matrix.buffer(i).fit.parameter_error.set([None] * len(self.inst.data.matrix.buffer(i).fit.parameter.get()))
            temp = []
            if i == firstbuffer:
                for val in header:
                    temp.append(val + spacer[:len(spacer) - len(val)])
                table += ''.join(temp) + "\n" + line + "\n"
            for j in range(len(self.inst.data.matrix.buffer(i).fit.parameter.get())):
                if j == 0:
                    buffer = str(i)
                else:
                    buffer = ''
                if str(self.inst.data.matrix.buffer(i).fit.free.get()[j]) == "True":
                    fixed = "False"
                else:
                    fixed = "True"
                lineout = [buffer, fitparams.paramid[j], str(self.inst.data.matrix.buffer(i).fit.parameter.get()[j]),
                           str(self.inst.data.matrix.buffer(i).fit.parameter_error.get()[j]),
                           str(self.inst.data.matrix.buffer(i).fit.link.get()[j]), fixed]
                temp = []
                for val in lineout:
                    temp.append(val + spacer[:len(spacer) - len(val)])
                table += ''.join(temp) + "\n"
        return table

    def do_der(self, *args):
        """\nCommand: calculate DERivative\n
               Description: Calculates the Derivative of a buffer.

               Example Usage:
               \tder 1 10   (Calculates the derivative of data in buffers 1 through 10)

               Default Input: N/A

               Default Options: N/A

               Options: N/A

               Notes: Derivatized data is appended to data matrix as new buffers
               """
        matrixlen = self.inst.data.matrix.length()
        inparse = inputprocessing.InputParser()
        inparse.prompt = ["First Buffer", "Last Buffer"]
        inparse.inputbounds = [[1, matrixlen], [1, matrixlen]]
        inparse.defaultinput = ['1', str(matrixlen)]
        if not inparse(args):
            return "Invalid input!"
        modifiers = inparse.modifiers
        userflags = inparse.cmdflags
        comparams = inparse.userinput

        min_buf = int(min(comparams))
        max_buf = int(max(comparams))
        for i in range(min_buf, max_buf+1, 1):
            new_buffer = self.inst.new_buffer()
            new_buffer.comments.set(f'{self.inst.data.matrix.buffer(i).comments.all_as_string()} Buffer{i} (derivative)')
            new_buffer.plot.series.name.set(f'{self.inst.data.matrix.buffer(i).plot.series.name.get()} (derivative)')
            new_buffer.plot.series.color.set(self.inst.data.matrix.buffer(i).plot.series.color.get())
            x, y = numericalmethods.calc_derivative(self.inst.data.matrix.buffer(i).data.x.get(),
                                                    self.inst.data.matrix.buffer(i).data.y.get())
            new_buffer.data.x.set(x)
            new_buffer.data.y.set(y)
            self.inst.data.matrix.add_buffer(new_buffer)

        return "\n Derivative Data for Buffer: {} through {} are written as Buffers {} through {}".format(*[min_buf, max_buf, matrixlen + 1, matrixlen + (max_buf-min_buf) + 1])

    def do_int(self, *args):
        """\nCommand: calculate running INTegral\n
               Description: Calculates the running integral of a buffer.

               Example Usage:
               \tint 1 10   (Calculates the running integral of data in buffers 1 through 10)

               Default Input: N/A

               Default Options: N/A

               Options: N/A

               Notes: Integral data is appended to data matrix as new buffers
               """
        matrixlen = self.inst.data.matrix.length()
        inparse = inputprocessing.InputParser()
        inparse.prompt = ["First Buffer", "Last Buffer"]
        inparse.inputbounds = [[1, matrixlen], [1, matrixlen]]
        inparse.defaultinput = ['1', str(matrixlen)]
        if not inparse(args):
            return "Invalid input!"
        modifiers = inparse.modifiers
        userflags = inparse.cmdflags
        comparams = inparse.userinput

        min_buf = int(min(comparams))
        max_buf = int(max(comparams))
        for i in range(min_buf, max_buf + 1, 1):
            new_buffer = self.inst.new_buffer()
            new_buffer.comments.set(f'{self.inst.data.matrix.buffer(i).comments.all_as_string()} Buffer{i} (derivative)')
            new_buffer.plot.series.name.set(f'{self.inst.data.matrix.buffer(i).plot.series.name.get()} (derivative)')
            new_buffer.plot.series.color.set(self.inst.data.matrix.buffer(i).plot.series.color.get())
            x, y = numericalmethods.calc_integral(self.inst.data.matrix.buffer(i).data.x.get(),
                                                  self.inst.data.matrix.buffer(i).data.y.get())
            new_buffer.data.x.set(x)
            new_buffer.data.y.set(y)
            self.inst.data.matrix.add_buffer(new_buffer)

        return "\n Integral Data for Buffer: {} through {} are written as Buffers {} through {}".format(
            *[min_buf, max_buf, matrixlen + 1, matrixlen + (max_buf - min_buf) + 1])

    def do_pbf(self, *args):
        """\nCommand: Print BuFfer\n
               Description: Displays all details of given buffer.

               Example Usage:
               \tpbf   (Displays all details of first buffer)
               \tpbf 1 (Displays all details of buffer 1)

               Default Input: 1

               Default Options: N/A

               Options: N/A
               """
        inparse = inputprocessing.InputParser()
        if not inparse(args):
            return "Invalid input!"
        userflags = inparse.cmdflags
        comparams = inparse.userinput

        if len(comparams) == 0:
            comparams.append(1)
        buf = int(comparams[0])
        if buf > self.inst.data.matrix.length():
            buf = self.inst.data.matrix.length()
        if buf < 1:
            buf = 1

        # code to print dictionary here:
        for key, value in self.inst.data.matrix.buffer(buf).to_dict().items():
            print(f'{str(key)}: {str(value)}')

        return "\n Showing Details of Buffer: " + str(buf)

    def help(self, cmd=None):
        def std_help():
            qc = '|'.join(self._quit_cmd)
            hc = '|'.join(self._help_cmd)
            res = '\n\tType [%s] command_name to get more help about particular command\n' % hc
            res += '\n\tType [%s] to quit program\n' % qc
            cl = [name[3:] for name in dir(self) if name.startswith('do_') and len(name) > 3]
            res += '\n\tAvailable commands: \n\t%s' % ('  '.join(sorted(cl))) + "\n"
            return res

        if not cmd:
            return std_help()
        else:
            try:
                fn = getattr(self, 'do_' + cmd)
                doc = fn.__doc__
                return doc or 'No documentation available for %s' % cmd
            except AttributeError:
                return std_help()

########################################################################################################################
########################################################################################################################
#############################                Command Related Methods Below                 #############################
########################################################################################################################
########################################################################################################################

    def updatefixlnk(self):
        if self.inst.data.plot_limits.is_active:
            firstbuffer, lastbuffer = self.inst.data.plot_limits.buffer_range.get()
        else:
            lastbuffer = self.inst.data.matrix.length()
            firstbuffer = 1
        for i in range(firstbuffer, lastbuffer+1):
            if len(self.inst.data.matrix.buffer(i).fit.link.get()) != len(self.inst.data.matrix.buffer(i).fit.parameter.get()):
                self.inst.data.matrix.buffer(i).fit.link.set([None] * len(self.inst.data.matrix.buffer(i).fit.parameter.get()))
                self.inst.data.matrix.buffer(i).fit.free.set([True] * len(self.inst.data.matrix.buffer(i).fit.parameter.get()))
        return True


    def runscript(self, scriptfile, interactivemode):
        user_input = None
        with open(scriptfile) as script:
            commandlist = script.readlines()
            for command in commandlist:
                if command[0] == "#" or command[0] == '\n':
                    continue
                elif command[0] == "!":
                    print(command[1:])
                    continue
                elif command[0] == "?":
                    user_input = input("\n" + command[1:] + "\nResponse: ")
                else:
                    if user_input is not None and "[user_input]" in command:
                        command.replace("[user_input]", str(user_input))
                    output = self(command)
                    if output and command.lower().strip() != 'quit' and command.lower().strip() != 'q':
                        print(output)
                        user_input = None
                    else:
                        sys.exit(0)
                if interactivemode:
                    interactiveinput = input("\nPress [q] to stop script, any other key to continue:  ")
                    if interactiveinput.lower() == "q":
                        break
        return "\nScript Executed Successfully!"




import matplotlib
try:
    matplotlib.use('QT5Agg')  # Required for gui plotting
except Exception as e:
    try:
        matplotlib.use('PyQT5')  # Required for gui plotting
    except Exception as e:
        matplotlib.use('Agg')  # Permits running headless
import matplotlib.pyplot as pl
from matplotlib.ticker import MaxNLocator
import matplotlib.patheffects as pe
from mpl_toolkits.axes_grid1 import make_axes_locatable
import numpy as np
from io import BytesIO as BIO
from PIL import Image
import warnings
warnings.filterwarnings("ignore")


class plotter(object):
    def __init__(self, data_instance):
        pl.close('all')
        self.__global_xlim = (-0.1, 0.1)
        self.__global_ylim = (-0.1, 0.1)
        self.__global_ylim_resid = (-0.1, 0.1)
        self.__resid_ax = False
        self.__scatter_ax = False
        self.__scatter_fig = False
        self.__plot = None
        self.__startplot = None
        self.inst = data_instance

    def __call__(self, pltarray, *args, **kwargs):
        print("Plotting...")
        pl.close('all')
        return_bytes = False
        scan = False
        auto = False
        delay = 0.5
        dpi = 50
        tight = True
        self.__global_xlim = (-0.1, 0.1)
        self.__global_ylim = (-0.1, 0.1)
        self.__global_ylim_resid = (-0.1, 0.1)
        self.__resid_ax = False
        self.__scatter_ax = False
        self.__scatter_fig = False
        self.__plot = None
        self.__startplot = None


        if 'get_bytes' in kwargs:
            return_bytes = bool(kwargs['get_bytes'])
        if 'scan' in kwargs:
            scan = kwargs['scan']
        if 'auto' in kwargs:
            auto = kwargs['auto']
        if 'delay' in kwargs:
            delay = kwargs['delay']
        if 'dpi' in kwargs:
            dpi = int(kwargs['dpi'])
        if 'tight' in kwargs:
            tight = int(kwargs['tight'])

        self.__set_plotting_backend(return_bytes)
        self.__get_global_lims(pltarray)

        if len(self.inst.matrix._Matrix__buffer_list) == 0:
            return "No data in matrix!"

        for i in range(len(pltarray)):
            buffer_number = pltarray[i]
            buffer = self.inst.matrix._Matrix__buffer_list[buffer_number-1]

            # determine plot type, and call correct routine
            plot_type = self.__get_plot_type(buffer)
            if plot_type == 'heatmap':
                self.__plot = self.__heatmap(buffer, *args, **kwargs)
            else:
                # scatter/line plots can be superimposed so plot has to be initialized outside of plotting method to permit stacking
                if i == 0:
                    self.__start_plot = pl.subplots(nrows=1, ncols=1, figsize=(8, 6), sharey=True, sharex=True)
                    # pl.close('all')
                self.__start_plot, self.__plot = self.__scatterline(buffer_number, plot_type, *args, **kwargs)
                if not auto:
                    print(f"Displaying Buffer {buffer_number}")

            # If scan mode, iterate through stack
            if scan:
                if auto:
                    print(f"Displaying Buffer {pltarray[i]} for {delay} seconds...")
                    matplotlib.pyplot.pause(delay)
                else:
                    userin = input("Press any key to continue, [q] to quit loop:  ")
                    if userin.lower() == 'q':
                        break
                if i < len(pltarray)-1:
                    pass
                    # pl.close('all')

        # if return bytes requested, return bytes
        if return_bytes:
            return self.__get_byte_png(self.__plot, dpi=dpi, tight=tight)
        return "Plotting Complete"

    def __get_plot_type(self, buffer):
        all_plots = ['scatter', 'line', 'heatmap']
        default = 'scatter'

        plot_type = buffer.plot.type.get()

        if plot_type.lower() in all_plots:
            return plot_type.lower()

        X = buffer.data.x.get()
        XE = buffer.data.xe.get()
        Y = buffer.data.y.get()
        YE = buffer.data.ye.get()
        Z = buffer.data.z.get()
        ZE = buffer.data.ze.get()
        X_model = buffer.model.x.get()
        Y_model = buffer.model.y.get()
        Z_model = buffer.model.z.get()
        X_cat = buffer.category.x.get()
        Y_cat = buffer.category.y.get()
        Z_cat = buffer.category.z.get()

        if len(X) > 1 and len(Y) > 1 and len(Z) == 0 and (len(X_model) == 0 or len(Y_model) == 0):
            return 'line'
        elif len(X_cat) > 1 and len(Y_cat) > 1 and len(Z) > 1:
            return 'heatmap'
        return default

    def __heatmap(self, buffer, *args, **kwargs):
        X = buffer.data.x.get()
        Y = buffer.data.y.get()
        Z = buffer.data.z.get()
        X_cat = buffer.category.x.get()
        Y_cat = buffer.category.y.get()
        X_title = buffer.plot.axis.x.title.get()
        X_title_size = buffer.plot.axis.x.label.size.get()
        Y_title = buffer.plot.axis.y.title.get()
        Y_title_size = buffer.plot.axis.y.label.size.get()
        Z_title = buffer.plot.axis.z.title.get()
        Z_title_size = buffer.plot.axis.z.label.size.get()
        plot_title = buffer.plot.title.get()

        X_axis = X if not X_cat else X_cat
        Y_axis = Y if not Y_cat else Y_cat

        data = Z.reshape((len(Y_axis), len(X_axis)))

        fig, ax = pl.subplots()
        pl.rcParams["font.size"] = min(X_title_size, Y_title_size)
        pl.gca().tick_params(axis='y', pad=8, labelsize=Y_title_size)
        pl.gca().tick_params(axis='x', pad=8, labelsize=X_title_size)

        # Plot the heatmap
        im = ax.imshow(data, cmap=matplotlib.cm.rainbow)

        # We want to show all ticks...
        ax.set_xticks(np.arange(data.shape[1]))
        ax.set_yticks(np.arange(data.shape[0]))
        # ... and label them with the respective list entries.
        ax.set_xticklabels(X_axis)
        ax.set_yticklabels(Y_axis)
        pl.xlabel(X_title, size=X_title_size)
        pl.ylabel(Y_title, size=Y_title_size)

        # Let the horizontal axes labeling appear on top.
        ax.tick_params(top=True, bottom=False,
                       labeltop=True, labelbottom=False)

        # Rotate the tick labels and set their alignment.
        pl.setp(ax.get_xticklabels(), rotation=-30, ha="right",
                rotation_mode="anchor")

        # Turn spines off and create white grid.
        for edge, spine in ax.spines.items():
            spine.set_visible(False)
        ax.set_xticks(np.arange(data.shape[1] + 1) - .5, minor=True)
        ax.set_yticks(np.arange(data.shape[0] + 1) - .5, minor=True)
        ax.grid(which="minor", color="w", linestyle='-', linewidth=2)
        ax.tick_params(which="minor", bottom=False, left=False)

        # Loop over data dimensions and create text annotations.
        for i in range(len(Y_axis)):
            for j in range(len(X_axis)):
                ax.text(j, i, data[i, j], ha="center", va="center", color='k')

        ax.set_title(plot_title)

        # Create colorbar
        # create an axes on the right side of ax. The width of cax will be 5%
        # of ax and the padding between cax and ax will be fixed at 0.1 inch.
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.1)
        cbar = pl.colorbar(im, cax=cax)
        cbar.ax.set_ylabel(Z_title, rotation=-90, va="bottom", labelpad=5)

        fig.canvas.flush_events()
        return pl

    def __scatterline(self, buffer_number, plot_type, *args, **kwargs):
        buffer = self.inst.matrix._Matrix__buffer_list[buffer_number-1]
        black_models = False
        scan = False

        if 'black_models' in kwargs:
            black_models = kwargs['black_models']
        if 'scan' in kwargs:
            scan = kwargs['scan']

        #####################################
        # declare variables to tidy up code #
        # ___________________________________#
        X = buffer.data.x.get()
        XE = buffer.data.xe.get()
        Y = buffer.data.y.get()
        YE = buffer.data.ye.get()
        Z = buffer.data.z.get()
        ZE = buffer.data.ze.get()
        show_data = buffer.data.is_visible
        X_model = buffer.model.x.get()
        Y_model = buffer.model.y.get()
        Z_model = buffer.model.z.get()
        model_color = buffer.model.color.get()
        show_model = buffer.model.is_visible
        X_resid = buffer.residuals.x.get()
        Y_resid = buffer.residuals.y.get()
        Z_resid = buffer.residuals.z.get()
        show_residuals = buffer.residuals.is_visible
        series_color = matplotlib.colors.to_hex(buffer.plot.series.color.get() if buffer.plot.series.color.get() != '' else 'r', keep_alpha=True)
        data_weight = buffer.plot.series.weight.get() if buffer.data.weight.get() <= 0 else buffer.data.weight.get()
        model_weight = data_weight/5 if buffer.model.weight.get() <= 0 else buffer.model.weight.get() 
        x_axis_title = buffer.plot.axis.x.title.get()
        y_axis_title = buffer.plot.axis.y.title.get()
        z_axis_title = buffer.plot.axis.z.title.get()
        x_title_size = buffer.plot.axis.x.label.size.get()
        y_title_size = buffer.plot.axis.y.label.size.get()
        z_title_size = buffer.plot.axis.z.label.size.get()
        x_title_show = buffer.plot.axis.x.label.is_visible
        y_title_show = buffer.plot.axis.y.label.is_visible
        z_title_show = buffer.plot.axis.z.label.is_visible
        show_x_lines = buffer.plot.axis.x.lines.is_visible
        x_line_color = buffer.plot.axis.x.lines.color.get()
        x_line_weight = buffer.plot.axis.x.lines.weight.get()
        x_line_is_outlined = buffer.plot.axis.x.lines.outline.is_visible
        x_line_style = buffer.plot.axis.x.lines.line_style.get()
        x_lines = buffer.plot.axis.x.lines.get()
        show_y_lines = buffer.plot.axis.y.lines.is_visible
        y_line_color = buffer.plot.axis.y.lines.color.get()
        y_line_weight = buffer.plot.axis.y.lines.weight.get()
        y_line_is_outlined = buffer.plot.axis.y.lines.outline.is_visible
        y_line_style = buffer.plot.axis.y.lines.line_style.get()
        y_lines = buffer.plot.axis.y.lines.get()
        show_peaks = buffer.plot.axis.y.peaks.is_visible
        peak_indicies = buffer.plot.axis.y.peaks.get()
        peaks = [Y[p] for p in peak_indicies]
        show_peak_bounds = buffer.plot.axis.y.peak_bounds.is_visible
        peak_bounds_indicies = buffer.plot.axis.y.peak_bounds.get()
        show_integrals = buffer.plot.axis.x.integrals.is_visible
        integral_indicies_pairs = buffer.plot.axis.x.integrals.get()
        show_polygons = buffer.plot.polygons.is_visible
        polygon_verticies = buffer.plot.polygons.get()
        x_axis_type = buffer.plot.axis.x.axis_scale.get()
        y_axis_type = buffer.plot.axis.y.axis_scale.get()
        z_axis_type = buffer.plot.axis.z.axis_scale.get()
        x_axis_range = buffer.plot.axis.x.range.get() if len(buffer.plot.axis.x.range.get()) > 1 else buffer.data.x.range()
        y_axis_range = buffer.plot.axis.y.range.get() if len(buffer.plot.axis.y.range.get()) > 1 else buffer.data.y.range()
        z_axis_range = buffer.plot.axis.z.range.get() if len(buffer.plot.axis.z.range.get()) > 1 else buffer.data.z.range()
        # ___________________________________#
        # END Local Variable Definitions    #
        #####################################
        if scan: # if scan create new figure each iteration
            pass
            # self.__start_plot = pl.subplots(nrows=1, ncols=1, figsize=(8, 6), sharey=True, sharex=True)
        if not self.__scatter_ax:
            self.__scatter_fig, self.__scatter_ax = self.__start_plot
            self.__scatter_ax.plot(1, 1)
            self.__scatter_ax.set_xscale(x_axis_type)
            self.__scatter_ax.set_yscale(y_axis_type)
            self.__scatter_ax.grid()
            pl.rcParams["font.size"] = 10
            pl.gca().tick_params(axis='y', pad=8, labelsize=y_title_size)
            pl.gca().tick_params(axis='x', pad=8, labelsize=x_title_size)
            if x_title_show:
                pl.xlabel(x_axis_title, size=x_title_size)
            if y_title_show:
                pl.ylabel(y_axis_title, size=y_title_size)

        if not scan and len(x_axis_range) > 0:
            # do nothing
            pass
        elif scan:
            padding = (max(X)-min(X)) * 0.00
            x_axis_range = (min(X) - padding, max(X) + padding)
        else:
            padding = (max(self.__global_xlim) - min(self.__global_xlim)) * 0.00
            x_axis_range = (min(self.__global_xlim) - padding, max(self.__global_xlim) + padding)

        if scan:
            padding = (max(Y) - min(Y)) * 0.05
            y_axis_range = (min(Y) - padding, max(Y) + padding)
            if len(Y_model) > 0:
                padding_resid = (max(Y_model) - min(Y_model)) * 0.05
                y_resid_axis_range = (min(Y_model) - padding_resid, max(Y_model) + padding_resid)
            else:
                y_resid_axis_range = (0,0)
        else:
            if len(y_axis_range) == 0:
                padding = (max(self.__global_ylim) - min(self.__global_ylim)) * 0.05
                y_axis_range = (min(self.__global_ylim) - padding, max(self.__global_ylim) + padding)
            padding_resid = (max(self.__global_ylim_resid) - min(self.__global_ylim_resid)) * 0.05
            y_resid_axis_range = (min(self.__global_ylim_resid) - padding_resid, max(self.__global_ylim_resid) + padding_resid)

            # Set Plot ranges
            pl.xlim(min(x_axis_range), max(x_axis_range))
            pl.ylim(min(y_axis_range), max(y_axis_range))

        # If Z data, treat as heatmap scatter
        if len(Z) == 0 or len(set(Z)) == 1:
            if plot_type == 'scatter':
                self.__scatter_ax.scatter(X, Y, color=series_color, s=data_weight * 4, zorder=0)  # weight is by area, correct to height val
            else:
                self.__scatter_ax.plot(X, Y, color=series_color, linestyle='-', linewidth=data_weight, zorder=0)
        else:
            # color points by z value and sort to not obscure data in plot
            sorted_X, sorted_Y, sorted_Z = self.__sort_by_z(X, Y, Z)
            if plot_type == 'scatter':
                self.__scatter_ax.scatter(sorted_X, sorted_Y, c=sorted_Z, s=data_weight * 4, cmap=matplotlib.cm.rainbow, zorder=0)  # weight is by area, correct to height val
            else:
                self.__scatter_ax.plot(sorted_X, sorted_Y, c=sorted_Z, linestyle='-', linewidth=data_weight, cmap=matplotlib.cm.rainbow, zorder=0)

        #####################################
        #            Model Plot             #
        # __________________________________#
        if len(X_model) > 0 and show_model:
            if black_models:
                self.__scatter_ax.plot(X_model, Y_model, linestyle='-', color='#000000', linewidth=model_weight, zorder=1)
            elif model_color == series_color:
                self.__scatter_ax.plot(X_model, Y_model, linestyle='-', color=series_color, linewidth=model_weight,
                        path_effects=[pe.Stroke(linewidth=model_weight *2, foreground='#000000'), pe.Normal()],
                        zorder=1)
            else:
                self.__scatter_ax.plot(X_model, Y_model, linestyle='-', color=model_color, linewidth=model_weight,
                                       zorder=1)
        #___________________________________#
        #        END Model Plot             #
        #####################################

        #####################################
        #            Residual Plot          #
        # __________________________________#
        if len(X_resid) > 0 and None not in (min(y_resid_axis_range), max(y_resid_axis_range)) and show_residuals:
            if not self.__resid_ax:
                divider = make_axes_locatable(self.__scatter_ax)
                self.__resid_ax = divider.append_axes("bottom", size="20%", pad=0)
                self.__scatter_ax.figure.add_axes(self.__resid_ax)
                self.__resid_ax.set_xscale(x_axis_type)
                self.__resid_ax.set_yscale(y_axis_type)
                # end residuals plot
                self.__scatter_ax.set_xticklabels([])
                pl.gca().yaxis.set_major_locator(MaxNLocator(prune='upper'))
                # Set x bounds to match data plot
                pl.xlim(min(x_axis_range), max(x_axis_range))
                pl.xlabel(x_axis_title, size=x_title_size)
                self.__resid_ax.axhline(y=0, linestyle='-', color='#000000', linewidth=2, zorder=0)
                self.__resid_ax.grid()
                pl.ylabel('Residuals')
            centered_range = (-1*max(abs(y_resid_axis_range[0]), abs(y_resid_axis_range[1])),
                                max(abs(y_resid_axis_range[0]), abs(y_resid_axis_range[1])))
            self.__resid_ax.set_ylim(min(centered_range), max(centered_range))
            self.__resid_ax.plot(X_resid, Y_resid, linestyle='-', color=series_color, linewidth=model_weight)
        #___________________________________#
        #        END Residual Plot          #
        #####################################

        #####################################
        #     X-line and Y-line Plotting    #
        #___________________________________#
        if show_x_lines and len(x_lines) >= 1:
            for line in x_lines:
                if x_line_is_outlined:
                    self.__scatter_ax.axvline(x=line, linestyle=x_line_style, color=x_line_color,
                                              linewidth=x_line_weight, zorder=2,
                                              path_effects=[
                                                  pe.Stroke(linewidth=x_line_weight * 2, foreground='#000000'),
                                                  pe.Normal()])
                else:
                    self.__scatter_ax.axvline(x=line, linestyle=x_line_style, color=x_line_color,
                                              linewidth=x_line_weight, zorder=2)
        if show_y_lines and len(y_lines) >= 1:
            for line in y_lines:
                if y_line_is_outlined:
                    self.__scatter_ax.axhline(y=line, linestyle=y_line_style, color=y_line_color,
                                              linewidth=y_line_weight, zorder=2,
                                              path_effects=[
                                                  pe.Stroke(linewidth=y_line_weight * 2, foreground='#000000'),
                                                  pe.Normal()])
                else:
                    self.__scatter_ax.axhline(y=line, linestyle=y_line_style, color=y_line_color,
                                              linewidth=y_line_weight, zorder=2)
        # ___________________________________#
        #  END X-line and Y-line Plotting   #
        #####################################

        #####################################
        #           Peak Plotting           #
        # ___________________________________#
        if show_peaks and len(peak_indicies) >= 1:
            peak_x = []
            peak_y = []
            for p in peak_indicies:
                peak_x.append(X[p])
                peak_y.append(Y[p])
            self.__scatter_ax.plot(peak_x, peak_y, "d", markerfacecolor='#808080', markeredgecolor='#000000', zorder=3)
        if show_peak_bounds and len(peak_bounds_indicies) >= 1:
            peak_bound_x = []
            peak_bound_y = []
            for p in peak_bounds_indicies:
                p_start = p[0]
                p_end = p[1]
                peak_bound_x.append(X[p_start])
                peak_bound_y.append(Y[p_start])
                peak_bound_x.append(X[p_end])
                peak_bound_y.append(Y[p_end])
            self.__scatter_ax.plot(peak_bound_x, peak_bound_y, "d", markerfacecolor='#808080', markeredgecolor='#000000', zorder=3)
        # ___________________________________#
        #       END Peak Plotting           #
        #####################################

        #####################################
        #         Integral  Plotting        #
        # ___________________________________#
        if show_integrals and len(integral_indicies_pairs) >= 1:
            if not isinstance(series_color, tuple):
                series_color = matplotlib.colors.to_rgba(series_color)
            int_color = list(series_color)
            int_color[3] = 0.4
            int_color = tuple(int_color)
            int_x = []
            int_y = []
            xy_vals = list(zip(X, Y))
            for int_indx in integral_indicies_pairs:
                int_start = (X[int_indx[0]], Y[int_indx[0]])
                int_end = (X[int_indx[-1]], Y[int_indx[-1]])
                verts = [(int_start[0], 0)] + xy_vals[int_indx[0]:int_indx[1]] + [(int_end[0], 0)]
                poly = matplotlib.patches.Polygon(verts, facecolor=int_color, edgecolor='#808080', zorder=0)
                self.__scatter_ax.add_patch(poly)
        # ___________________________________#
        #     END  Integral  Plotting       #
        #####################################

        #####################################
        #        Polygon Plotting           #
        # ___________________________________#
        if show_polygons and len(polygon_verticies) >= 1:
            for polygon in polygon_verticies:
                poly_x, poly_y = zip(*polygon)
                self.__scatter_ax.plot(poly_x, poly_y, linestyle='-', color='#808080', linewidth=data_weight, zorder=4,
                        path_effects=[pe.Stroke(linewidth=data_weight * 2, foreground='k'), pe.Normal()])
        # ___________________________________#
        #       END  Polygon Plotting       #
        #####################################

        self.__scatter_fig.canvas.flush_events()
        return [self.__scatter_fig, self.__scatter_ax], pl

    def __get_byte_png(self, pl, dpi=50, tight=True):
        global is_savefig
        BIOstream = BIO()
        is_savefig = True
        pl.savefig(BIOstream, format='png', dpi=dpi, bbox_inches='tight')
        pl.close('all')
        is_savefig = False
        BIOstream.seek(0)
        im = Image.open(BIOstream)
        im2 = im.convert('RGB').convert('P', palette=Image.ADAPTIVE)
        im2 = im2.resize(im2.size, Image.ANTIALIAS)
        BIOstream = BIO()
        im2.save(BIOstream, format='PNG', quality=95)
        BIOstream.seek(0)
        byte_png = BIOstream.getvalue()
        pl.close('all')
        return byte_png

    def __sort_by_z(self, X, Y, Z):
        idx = Z.argsort()
        sorted_x, sorted_y, sorted_z = X[idx], Y[idx], Z[idx]
        return sorted_x, sorted_y, sorted_z

    def __get_global_lims(self, pltarray):
        xmin = np.inf
        xmax = -np.inf
        ymin = np.inf
        ymax = -np.inf
        ymin_res = np.inf
        ymax_res = -np.inf

        for num in pltarray:
            if int(num) > len(self.inst.matrix._Matrix__buffer_list):
                continue
            buffer = self.inst.matrix._Matrix__buffer_list[num-1]
            if self.inst.matrix.buffer(num).data.x.length() > 0:
                if np.isfinite(buffer.data.x.min()) and buffer.data.x.min() <= xmin:
                    xmin = buffer.data.x.min()
                if np.isfinite(buffer.data.x.max()) and buffer.data.x.max() >= xmax:
                    xmax = buffer.data.x.max()
            if buffer.data.y.length() > 0:
                if np.isfinite(buffer.data.y.min()) and buffer.data.y.min() <= ymin:
                    ymin = buffer.data.y.min()
                if np.isfinite(buffer.data.y.max()) and buffer.data.y.max() >= ymax:
                    ymax = buffer.data.y.max()
            if buffer.residuals.y.length() > 0:
                if np.isfinite(buffer.residuals.y.min()) and buffer.residuals.y.min() <= ymin_res:
                    ymin_res = buffer.residuals.y.min()
                if np.isfinite(buffer.residuals.y.max()) and buffer.residuals.y.max() >= ymax_res:
                    ymax_res = buffer.residuals.y.max()

        self.__global_xlim = (xmin, xmax)
        self.__global_ylim = (ymin, ymax)
        self.__global_ylim_resid = (ymin_res, ymax_res)
        return True

    def __set_plotting_backend(self, return_bytes):
        # Agg permits headless figure production
        init_backend = matplotlib.rcParams['backend']

        if init_backend.lower() not in ['agg', 'qt5agg']:
            raise ValueError(f'Current MatplotLib backend is not supported: {init_backend}')

        if not return_bytes and init_backend != 'qt5agg':
            try:
                # not returning bytes try to enforce gui if backend is available.  Turn interactive mode on
                matplotlib.use('qt5agg')
                pl.ion()
            except Exception as e:
                print('Using non-GUI backend: ' + str(e))
                matplotlib.use('agg')
                pl.ioff()
        elif return_bytes and init_backend != 'agg':
            try:
                # returning bytes try to enforce NO gui if backend is available.  Turn interactive mode off
                matplotlib.use('agg')
                pl.ioff()
            except Exception as e:
                print('Using GUI backend: ' + str(e))
                matplotlib.use('qt5agg')
                pl.ion()

        pl.close('all')
        print(f"Using MatPlotLib backend: {str(matplotlib.rcParams['backend'])}")
        return
